from typing import Optional, Tuple

from .. import func
from ... import config
from ...param import forward


class CMOSMixin:
    """
    A mixin class for CMOS image processing. This class can be used to add
    CMOS-specific functionality to image processing classes.
    """

    def __init__(
        self,
        *args,
        subpixel_loc: Tuple[float, float] = (0, 0),
        subpixel_scale: float = 1.0,
        filename: Optional[str] = None,
        **kwargs,
    ):
        super().__init__(*args, filename=filename, **kwargs)
        if filename is not None:
            return
        self.subpixel_loc = subpixel_loc
        self.subpixel_scale = subpixel_scale

    @property
    def base_scale(self):
        """Get the base scale of the image, which is the subpixel scale."""
        return self.subpixel_scale

    @forward
    def pixel_collecting_area(self, I_, J_, upsample):
        # CMOS pixels only sensitive in sub area, so scale the pixel collecting area
        return self.pixel_area * self.subpixel_scale**2 / upsample**2

    def pixel_center_meshgrid(self, window=None, pad=0, upsample=1):
        """Get a meshgrid of pixel coordinates in the image, centered on the pixel grid."""
        if window is None:
            window = self.window
        return func.cmos_pixel_center_meshgrid(
            window.extent, pad, upsample, self.subpixel_loc, config.DTYPE, config.DEVICE
        )

    def copy(self, **kwargs):
        return super().copy(
            subpixel_loc=self.subpixel_loc, subpixel_scale=self.subpixel_scale, **kwargs
        )

    def fits_info(self):
        info = super().fits_info()
        info["SPIXLOC1"] = self.subpixel_loc[0]
        info["SPIXLOC2"] = self.subpixel_loc[1]
        info["SPIXSCL"] = self.subpixel_scale
        return info

    def load(self, filename: str, hduext: int = 0):
        hdulist = super().load(filename, hduext=hduext)
        if "SPIXLOC1" in hdulist[hduext].header:
            self.subpixel_loc = (
                hdulist[0].header.get("SPIXLOC1", 0),
                hdulist[0].header.get("SPIXLOC2", 0),
            )
        if "SPIXSCL" in hdulist[hduext].header:
            self.subpixel_scale = hdulist[0].header.get("SPIXSCL", 1.0)
        return hdulist



import numpy as np
from math import prod
from caskade import (
    Module as CModule,
    ActiveStateError,
    ParamConfigurationError,
    FillParamsArrayError,
)
from ..backend_obj import backend


class Module(CModule):

    def build_params_array_identities(self):
        identities = []
        for param in self.dynamic_params:
            numel = max(1, np.prod(param.batch_shape + param.shape))
            for i in range(numel):
                identities.append(f"{id(param)}_{i}")
        return identities

    def build_params_array_uncertainty(self):
        uncertainties = []
        for param in self.dynamic_params:
            if param.uncertainty is None:
                uncertainties.append(backend.zeros_like(param.value.flatten()))
            else:
                uncertainties.append(param.uncertainty.flatten())
        return backend.concatenate(tuple(uncertainties), dim=-1)

    def build_params_array_names(self):
        names = []
        for param in self.dynamic_params:
            numel = max(1, np.prod(param.batch_shape + param.shape))
            if numel == 1:
                names.append(param.name)
            else:
                for i in range(numel):
                    names.append(f"{param.name}_{i}")
        return names

    def build_params_array_units(self):
        units = []
        for param in self.dynamic_params:
            numel = max(1, np.prod(param.batch_shape + param.shape))
            for _ in range(numel):
                units.append(param.units)
        return units



from typing import Union, Optional

import numpy as np
from astropy.io import fits

from ...utils.initialize import auto_variance
from ... import config
from ...backend_obj import backend, ArrayLike
from ...errors import SpecificationConflict
from ..image_object import Image
from ..window import Window


class DataMixin:
    """Mixin for data handling in image objects.

    This mixin provides functionality for handling variance and mask,
    as well as other ancillary data.

    **Args:**
    - `mask`: A boolean mask indicating which pixels to ignore.
    - `std`: Standard deviation of the image pixels.
    - `variance`: Variance of the image pixels.
    - `weight`: Weights for the image pixels.

    Note that only one of `std`, `variance`, or `weight` should be
    provided at a time. If multiple are provided, an error will be raised.
    """

    def __init__(
        self,
        *args,
        mask: Optional[ArrayLike] = None,
        std: Optional[ArrayLike] = None,
        variance: Optional[ArrayLike] = None,
        weight: Optional[ArrayLike] = None,
        _mask: Optional[ArrayLike] = None,
        _weight: Optional[ArrayLike] = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        if _mask is None:
            self.mask = mask
        else:
            self._mask = _mask
        if (std is not None) + (variance is not None) + (weight is not None) > 1:
            raise SpecificationConflict(
                "Can only define one of: std, variance, or weight for a given image."
            )

        if _weight is not None:
            self._weight = _weight
        elif std is not None:
            self.std = std
        elif variance is not None:
            self.variance = variance
        else:
            self.weight = weight

        # Set nan pixels to be masked automatically
        self._mask = self._mask | backend.isnan(self._data)
        self._data = backend.nan_to_num(self._data, nan=0.0)

    @property
    def std(self):
        """Stores the standard deviation of the image pixels. This represents
        the uncertainty in each pixel value. It should always have the
        same shape as the image data. In the case where the standard
        deviation is not known, a tensor of ones will be created to
        stand in as the standard deviation values.

        The standard deviation is not stored directly, instead it is
        computed as $\\sqrt{1/W}$ where $W$ is the weights.

        """
        return backend.sqrt(self.variance)

    @std.setter
    def std(self, std):
        if std is None:
            self._weight = None
            return
        if isinstance(std, str) and std == "auto":
            self.weight = "auto"
            return
        self.weight = 1 / std**2

    @property
    def variance(self):
        """Stores the variance of the image pixels. This represents the
        uncertainty in each pixel value. It should always have the
        same shape as the image data. In the case where the variance
        is not known, a tensor of ones will be created to stand in as
        the variance values.

        The variance is not stored directly, instead it is
        computed as $\\frac{1}{W}$ where $W$ is the
        weights.

        """
        return backend.where(self.weight == 0, backend.inf, 1 / self.weight)

    @property
    def _variance(self):
        return backend.where(self._weight == 0, backend.inf, 1 / self._weight)

    @variance.setter
    def variance(self, variance):
        if variance is None:
            self._weight = None
            return
        if isinstance(variance, str) and variance == "auto":
            self.weight = "auto"
            return
        self.weight = 1 / variance

    @property
    def weight(self):
        """Stores the weight of the image pixels. This represents the
        uncertainty in each pixel value. It should always have the
        same shape as the image data. In the case where the weight
        is not known, a tensor of ones will be created to stand in as
        the weight values.

        The weights are used to proprtionately scale residuals in the
        likelihood. Most commonly this shows up as a :math:`\\chi^2`
        like:

        $$\\chi^2 = (\\vec{y} - \\vec{f(\\theta)})^TW(\\vec{y} - \\vec{f(\\theta)})$$

        which can be optimized to find parameter values. Using the
        Jacobian, which in this case is the derivative of every pixel
        wrt every parameter, the weight matrix also appears in the
        gradient:

        $$\\vec{g} = J^TW(\\vec{y} - \\vec{f(\\theta)})$$

        and the hessian approximation used in Levenberg-Marquardt:

        $$H \\approx J^TWJ$$

        """
        return backend.transpose(self._weight, 1, 0)

    @weight.setter
    def weight(self, weight):
        if weight is None:
            self._weight = backend.ones_like(self._data)
            return
        if isinstance(weight, str) and weight == "auto":
            weight = 1 / auto_variance(self.data, self.mask)
        self._weight = backend.transpose(
            backend.as_array(weight, dtype=config.DTYPE, device=config.DEVICE), 1, 0
        )
        if self._weight.shape != self._data.shape:
            raise SpecificationConflict(
                f"weight/variance must have same shape as data ({weight.shape} vs {self.data.shape})"
            )

    @property
    def _weight(self):
        return self.__weight

    @_weight.setter
    def _weight(self, value):
        if value is None:
            value = backend.ones_like(self._data)
        self.__weight = value

    @property
    def mask(self):
        """The mask stores a tensor of boolean values which indicate any
        pixels to be ignored. These pixels will be skipped in
        likelihood evaluations and in parameter optimization. It is
        common practice to mask pixels with pathological values such
        as due to cosmic rays or satellites passing through the image.

        In a mask, a True value indicates that the pixel is masked and
        should be ignored. False indicates a normal pixel which will
        inter into most calculations.

        If no mask is provided, all pixels are assumed valid.

        """
        return backend.transpose(self._mask, 1, 0)

    @mask.setter
    def mask(self, mask):
        if mask is None:
            self._mask = backend.zeros_like(self._data, dtype=backend.bool)
            return
        self._mask = backend.transpose(
            backend.as_array(mask, dtype=backend.bool, device=config.DEVICE), 1, 0
        )
        if self._mask.shape != self._data.shape:
            raise SpecificationConflict(
                f"mask must have same shape as data ({mask.shape} vs {self.data.shape})"
            )

    @property
    def _mask(self):
        return self.__mask

    @_mask.setter
    def _mask(self, value):
        if value is None:
            value = backend.zeros_like(self._data, dtype=backend.bool)
        self.__mask = value

    def to(self, dtype=None, device=None):
        """Converts the stored `Target_Image` data, variance, psf, etc to a
        given data type and device.

        """
        if dtype is not None:
            dtype = config.DTYPE
        if device is not None:
            device = config.DEVICE
        super().to(dtype=dtype, device=device)

        self._weight = backend.to(self._weight, dtype=dtype, device=device)
        self._mask = backend.to(self._mask, dtype=backend.bool, device=device)
        return self

    def copy_kwargs(self, **kwargs):
        """Produce a copy of this image with all of the same properties. This
        can be used when one wishes to make temporary modifications to
        an image and then will want the original again.

        """
        kwargs = {"_mask": self._mask, "_weight": self._weight, **kwargs}
        return getattr(super(), "copy_kwargs", lambda **kw: kw)(**kwargs)

    def get_window(self, other: Union[Image, Window], indices=None, **kwargs):
        """Get a sub-region of the image as defined by an other image on the sky."""
        if indices is None:
            indices = self.get_indices(other if isinstance(other, Window) else other.window)
        return super().get_window(
            other,
            _weight=self._weight[indices],
            _mask=self._mask[indices],
            indices=indices,
            **kwargs,
        )

    def fits_images(self):
        images = super().fits_images()
        images.append(fits.ImageHDU(backend.to_numpy(self.weight), name="WEIGHT"))
        images.append(
            fits.ImageHDU(
                backend.to_numpy(self.mask).astype(int),
                name="MASK",
            )
        )
        return images

    def load(self, filename: str, hduext: int = 0):
        """Load the image from a FITS file. This will load the data, WCS, and
        any ancillary data such as variance, mask, and PSF.

        """
        hdulist = super().load(filename, hduext=hduext)
        if "WEIGHT" in hdulist:
            self.weight = np.array(hdulist["WEIGHT"].data, dtype=np.float64)
        if "MASK" in hdulist:
            self.mask = np.array(hdulist["MASK"].data, dtype=bool)
        elif "DQ" in hdulist:
            self.mask = np.array(hdulist["DQ"].data, dtype=bool)
        return hdulist

    def reduce(self, scale: int, **kwargs) -> Image:
        """Returns a new `TargetImage` object with a reduced resolution
        compared to the current image. `scale` should be an integer
        indicating how much to reduce the resolution. If the
        `TargetImage` was originally (48,48) pixels across with a
        pixelscale of 1 and `reduce(2)` is called then the image will
        be (24,24) pixels and the pixelscale will be 2. If `reduce(3)`
        is called then the returned image will be (16,16) pixels
        across and the pixelscale will be 3.

        """
        MS = self._data.shape[0] // scale
        NS = self._data.shape[1] // scale

        return super().reduce(
            scale=scale,
            _weight=(
                1
                / backend.sum(
                    self._variance[: MS * scale, : NS * scale].reshape(MS, scale, NS, scale),
                    dim=(1, 3),
                )
            ),
            _mask=(
                backend.max(
                    self._mask[: MS * scale, : NS * scale].reshape(MS, scale, NS, scale), dim=(1, 3)
                )
            ),
            **kwargs,
        )



# Metropolis-Hasting Markov-Chain Monte-Carlo
from typing import Optional, Sequence

import numpy as np

try:
    import emcee
except ImportError:
    emcee = None

from .base import BaseOptimizer
from ..models import Model
from .. import config
from ..backend_obj import backend

__all__ = ("MHMCMC",)


class MHMCMC(BaseOptimizer):
    """Metropolis-Hastings Markov-Chain Monte-Carlo sampler, based on:
    https://en.wikipedia.org/wiki/Metropolis-Hastings_algorithm . This is simply
    a thin wrapper for the Emcee package, which is a well-known MCMC sampler.

    Note that the Emcee sampler requires multiple walkers to sample the
    parameter space efficiently. The number of walkers is set to twice the
    number of parameters by default, but can be made higher (not lower) if desired.
    This is done by passing a 2D array of shape (nwalkers, ndim) to the `fit` method.

    **Args:**
    -  `likelihood`: The likelihood function to use for the MCMC sampling. Can be "gaussian" or "poisson". Default is "gaussian".
    """

    def __init__(
        self,
        model: Model,
        initial_state: Optional[Sequence] = None,
        max_iter: int = 1000,
        likelihood="gaussian",
        **kwargs,
    ):
        super().__init__(model, initial_state, max_iter=max_iter, **kwargs)

        if emcee is None:
            raise ImportError(
                "The emcee package is required for MHMCMC sampling. Please install it with `pip install emcee` or the like."
            )
        self.likelihood = likelihood

        self.chain = []

    def density(self):
        """
        Returns the density of the model at the given state vector.
        This is used to calculate the likelihood of the model at the given state.
        """
        if self.likelihood == "gaussian":
            vll = backend.vmap(self.model.gaussian_log_likelihood)
        elif self.likelihood == "poisson":
            vll = backend.vmap(self.model.poisson_log_likelihood)
        else:
            raise ValueError(f"Unknown likelihood type: {self.likelihood}")

        def dens(state: np.ndarray) -> np.ndarray:
            state = backend.as_array(state, dtype=config.DTYPE, device=config.DEVICE)
            return backend.to_numpy(vll(state))

        return dens

    def fit(
        self,
        state: Optional[np.ndarray] = None,
        nsamples: Optional[int] = None,
        restart_chain: bool = True,
        skip_initial_state_check: bool = True,
        flat_chain: bool = True,
    ):
        """
        Performs the MCMC sampling using a Metropolis Hastings acceptance step and records the chain for later examination.
        """

        if nsamples is None:
            nsamples = self.max_iter

        if state is None:
            state = self.current_state

        if len(state.shape) == 1:
            nwalkers = state.shape[0] * 2
            state = state * np.random.normal(loc=1, scale=0.01, size=(nwalkers, state.shape[0]))
        else:
            nwalkers = state.shape[0]
        ndim = state.shape[1]
        sampler = emcee.EnsembleSampler(nwalkers, ndim, self.density(), vectorize=True)
        state = sampler.run_mcmc(state, nsamples, skip_initial_state_check=skip_initial_state_check)
        if restart_chain:
            self.chain = sampler.get_chain(flat=flat_chain)
        else:
            self.chain = np.append(self.chain, sampler.get_chain(flat=flat_chain), axis=0)
        self.model.set_values(
            backend.as_array(self.chain[-1], dtype=config.DTYPE, device=config.DEVICE)
        )
        return self



from ...backend_obj import ArrayLike


def all_subclasses(cls):
    return set(cls.__subclasses__()).union(
        [s for c in cls.__subclasses__() for s in all_subclasses(c)]
    )


def downsample(img: ArrayLike, scale: int = 1):
    if scale == 1:
        return img
    MS = img.shape[0] // scale
    NS = img.shape[1] // scale

    return img[: MS * scale, : NS * scale].reshape(MS, scale, NS, scale).sum(axis=(1, 3))


def downsample_mean(img: ArrayLike, scale: int = 1):
    if scale == 1:
        return img
    MS = img.shape[0] // scale
    NS = img.shape[1] // scale

    return img[: MS * scale, : NS * scale].reshape(MS, scale, NS, scale).mean(axis=(1, 3))



from .lm import lm_step, hessian, gradient, hessian_poisson, gradient_poisson
from .slalom import slalom_step
from .mala import mala

__all__ = [
    "lm_step",
    "hessian",
    "gradient",
    "slalom_step",
    "hessian_poisson",
    "gradient_poisson",
    "mala",
]



from .group_model_object import GroupModel
from ..image import PSFImage
from ..errors import InvalidTarget
from ..param import forward

__all__ = ["PSFGroupModel"]


class PSFGroupModel(GroupModel):
    """
    A group of PSF models. Behaves similarly to a `GroupModel`, but specifically designed for PSF models.
    """

    _model_type = "psf"
    usable = True
    normalize_psf = True

    _options = ("normalize_psf",)

    @property
    def target(self):
        try:
            return self._target
        except AttributeError:
            return None

    @target.setter
    def target(self, target):
        if not (target is None or isinstance(target, PSFImage)):
            raise InvalidTarget("GroupModel target must be a PSFImage instance.")
        try:
            del self._target  # Remove old target if it exists
        except AttributeError:
            pass

        self._target = target

    @forward
    def sample(self, *args, **kwargs):
        """Sample the PSF group model on the target image."""
        image = self.target.model_image(self.window)

        for model in self.models:
            model_image = model()
            self._ensure_vmap_compatible(image, model_image)
            image += model_image

        if self.normalize_psf:
            image.normalize()
        return image



import matplotlib
import matplotlib.pyplot as plt
import pytest
import numpy as np
import astrophot as ap


@pytest.fixture(autouse=True)
def no_block_show(monkeypatch):
    def close_show(*args, **kwargs):
        # plt.savefig("/dev/null")  # or do nothing
        plt.close("all")

    monkeypatch.setattr(plt, "show", close_show)

    # Also ensure we are in a non-GUI backend
    matplotlib.use("Agg")


@pytest.fixture()
def sersic(request):

    if not hasattr(request, "param"):
        request.param = {}
    np.random.seed(request.param.get("seed", 12345))
    shape = request.param.get("shape", (52, 50))
    mask = request.param.get("mask", None)
    if mask is None:
        mask = np.zeros(shape, dtype=bool)
        mask[0][0] = True
    target = request.param.get("target", None)
    pixelscale = 0.8
    if target is None:
        target = ap.TargetImage(
            data=np.zeros(shape),
            pixelscale=pixelscale,
            psf=ap.utils.initialize.gaussian_psf(2 / pixelscale, 11, pixelscale),
            mask=mask,
            zeropoint=21.5,
        )

    MODEL = ap.models.SersicGalaxy(
        name="basic_sersic_model",
        target=target,
        center=request.param.get("center", [20.5, 21.4]),
        PA=request.param.get("PA", 45 * np.pi / 180),
        q=request.param.get("q", 0.7),
        n=request.param.get("n", 1.5),
        Re=request.param.get("Re", 15.1),
        Ie=request.param.get("Ie", 10.0),
        sampling_mode="quad:5",
    )

    if request.param.get("target", None) is None:
        img = ap.backend.to_numpy(MODEL().data)
        target.data = (
            img
            + np.random.normal(scale=0.5, size=img.shape)
            + np.random.normal(scale=np.sqrt(img) / 10)
        )
        target.variance = 0.5**2 + img / 100

    return MODEL



from typing import Optional, Literal

import numpy as np

from ...param import forward
from ...backend_obj import backend, ArrayLike
from ... import config
from ...image import JacobianImage
from .. import func
from ...errors import SpecificationConflict
from ...utils.integration import quad_table


class SampleMixin:
    """
    **Options:**
    -    `sampling_mode`: The method used to sample the model in image pixels. Options are:
            - `auto`: Automatically choose the sampling method based on the image size.
            - `midpoint`: Use midpoint sampling, evaluate the brightness at the center of each pixel.
            - `simpsons`: Use Simpson's rule for sampling integrating each pixel.
            - `quad:x`: Use quadrature sampling with order x, where x is a positive integer to integrate each pixel.
    -    `integrate_mode`: The method used to select pixels to integrate further where the model varies significantly. Options are:
            - `none`: No extra integration is performed (beyond the sampling_mode).
            - `bright`: Select the brightest pixels for further integration.
            - `threshold`: Select pixels which show signs of significant higher order derivatives.
    -    `integrate_tolerance`: The tolerance for selecting a pixel in the integration method. This is the total flux fraction that is integrated over the image.
    -    `integrate_fraction`: The fraction of the pixels to super sample during integration.
    -    `integrate_max_depth`: The maximum depth of the integration method.
    -    `integrate_gridding`: The gridding used for the integration method to super-sample a pixel at each iteration.
    -    `integrate_quad_order`: The order of the quadrature used for the integration method on the super sampled pixels.
    """

    integrate_fraction = 0.05  # fraction of the pixels to super sample
    integrate_max_depth = 2
    integrate_gridding = 5
    integrate_quad_order = 3

    _options = (
        "sampling_mode",
        "integrate_mode",
        "integrate_fraction",
        "integrate_max_depth",
        "integrate_gridding",
        "integrate_quad_order",
    )

    def __init__(self, *args, sampling_mode="auto", integrate_mode="bright", **kwargs):
        super().__init__(*args, **kwargs)
        self.sampling_mode = sampling_mode
        self.integrate_mode = integrate_mode

    @forward
    def _bright_integrate(
        self,
        sample: ArrayLike,
        i: ArrayLike,
        j: ArrayLike,
        upsample: int,
        pixel_brightness: callable,
    ) -> ArrayLike:
        sample = func.bright_integrate(
            sample,
            i,
            j,
            pixel_brightness,
            scale=self.target.base_scale / upsample,
            bright_frac=self.integrate_fraction,
            quad_order=self.integrate_quad_order,
            gridding=self.integrate_gridding,
            max_depth=self.integrate_max_depth,
        )
        return sample

    @forward
    def _curvature_integrate(
        self,
        sample: ArrayLike,
        i: ArrayLike,
        j: ArrayLike,
        upsample: int,
        pixel_brightness: callable,
    ) -> ArrayLike:
        kernel = func.curvature_kernel(config.DTYPE, config.DEVICE)
        curvature = (
            backend.abs(
                backend.pad(
                    backend.conv2d(
                        sample.reshape(1, 1, *sample.shape),
                        kernel.reshape(1, 1, *kernel.shape),
                        padding="valid",
                    ),
                    (0, 0, 0, 0, 1, 1, 1, 1),
                    mode="replicate",
                )
            )
            .squeeze(0)
            .squeeze(0)
        )
        N = max(1, int(np.prod(i.shape) * self.integrate_fraction))
        select = backend.topk(curvature.flatten(), N)[1]

        sample_flat = sample.flatten()
        sample_flat = backend.fill_at_indices(
            sample_flat,
            select,
            func.recursive_quad_integrate(
                i.flatten()[select],
                j.flatten()[select],
                pixel_brightness,
                scale=self.target.base_scale / upsample,
                curve_frac=self.integrate_fraction,
                quad_order=self.integrate_quad_order,
                gridding=self.integrate_gridding,
                max_depth=self.integrate_max_depth,
            ),
        )
        return sample_flat.reshape(sample.shape)

    @property
    def sampling_mode(self):
        return self._sampling_mode

    @sampling_mode.setter
    def sampling_mode(self, sampling_mode):
        if sampling_mode == "auto":
            sampling_mode = "midpoint"
            try:
                N = np.prod(self.window.shape)
                if N <= 100:
                    sampling_mode = "quad:5"
                elif N <= 10000:
                    sampling_mode = "simpsons"
            except:
                pass
        if sampling_mode == "midpoint":
            self._pixel_meshgridder = lambda im, w, p, u: im.pixel_center_meshgrid(w, p, u)
            self._pixel_integrator = func.pixel_center_integrator
            self._pixel_center_finder = lambda i, j: (i, j)
        elif sampling_mode == "simpsons":
            self._pixel_meshgridder = lambda im, w, p, u: im.pixel_simpsons_meshgrid(w, p, u)
            self._pixel_integrator = func.pixel_simpsons_integrator
            self._pixel_center_finder = lambda i, j: (i[1:-1:2, 1:-1:2], j[1:-1:2, 1:-1:2])
        elif sampling_mode.startswith("quad:"):
            order = int(sampling_mode.split(":")[1])
            self._pixel_meshgridder = lambda im, w, p, u: im.pixel_quad_meshgrid(
                w, p, u, order=order
            )[:2]
            _, _, w = quad_table(order, config.DTYPE, config.DEVICE)
            w = w.flatten()
            self._pixel_integrator = lambda z: func.pixel_quad_integrator(z, w)
            self._pixel_center_finder = lambda i, j: (i[..., order**2 // 2], j[..., order**2 // 2])
        elif sampling_mode.startswith("upsample:"):
            upsample = int(sampling_mode.split(":")[1])
            if upsample % 2 != 1:
                raise SpecificationConflict(
                    f"Upsample factor for 'sample_mode' must be an odd integer, got {upsample} for model {self.name}"
                )
            self._pixel_meshgridder = lambda im, w, p, u: im.pixel_center_meshgrid(
                w, upsample * p, upsample * u
            )
            self._pixel_integrator = lambda z: func.downsample_mean(z, upsample)
            self._pixel_center_finder = lambda i, j: (
                i[upsample // 2 :: upsample, upsample // 2 :: upsample],
                j[upsample // 2 :: upsample, upsample // 2 :: upsample],
            )
        else:
            raise SpecificationConflict(
                f"Unknown sampling mode {sampling_mode} for model {self.name}"
            )
        self._sampling_mode = sampling_mode

    @property
    def integrate_mode(self):
        return self._integrate_mode

    @integrate_mode.setter
    def integrate_mode(self, integrate_mode):
        if integrate_mode == "bright":
            self._adaptive_integrator = self._bright_integrate
        elif integrate_mode == "curvature":
            self._adaptive_integrator = self._curvature_integrate
        elif integrate_mode == "none":
            self._adaptive_integrator = lambda z, *a, **kw: z
        else:
            raise SpecificationConflict(
                f"Unknown integrate mode {integrate_mode} for model {self.name}"
            )
        self._integrate_mode = integrate_mode


class GradMixin:
    """
    **Options:**
    -    `jacobian_maxparams`: The maximum number of parameters before the Jacobian will be broken into smaller chunks. This is helpful for limiting the memory requirements to fit a model.
    """

    # Maximum size of parameter list before jacobian will be broken into smaller chunks, this is helpful for limiting the memory requirements to build a model, lower jacobian_chunksize is slower but uses less memory
    jacobian_maxparams = 10

    _options = ("jacobian_maxparams",)

    def _jacobian(
        self,
        params_pre: ArrayLike,
        params: ArrayLike,
        params_post: ArrayLike,
    ) -> ArrayLike:
        # return jacfwd( # this should be more efficient, but the trace overhead is too high
        #     lambda x: self.sample(
        #         window=window, params=torch.cat((params_pre, x, params_post), dim=-1)
        #     ).data
        # )(params)
        return backend.jacobian(
            lambda x: self(params=backend.concatenate((params_pre, x, params_post), dim=-1))._data,
            params,
        )

    def jacobian(
        self,
        pass_jacobian: Optional[JacobianImage] = None,
        params: Optional[ArrayLike] = None,
    ) -> JacobianImage:

        if pass_jacobian is None:
            jac_img = self.target[self.window].jacobian_image(
                parameters=self.build_params_array_identities()
            )
        else:
            jac_img = pass_jacobian

        # No dynamic params
        if params is None:
            params = self.get_values()
        if len(params.shape) == 0 or params.shape[-1] == 0:
            return jac_img

        identities = self.build_params_array_identities()
        if len(jac_img.match_parameters(identities)[0]) == 0:
            return jac_img

        target = self.target[self.window]
        if len(params) > self.jacobian_maxparams:  # handle large number of parameters
            chunksize = len(params) // self.jacobian_maxparams + 1
            for i in range(0, len(params), chunksize):
                params_pre = params[:i]
                params_chunk = params[i : i + chunksize]
                params_post = params[i + chunksize :]
                jac_chunk = self._jacobian(params_pre, params_chunk, params_post)
                jac_img += target.jacobian_image(
                    parameters=identities[i : i + chunksize],
                    data=jac_chunk,
                )
        else:
            jac = self._jacobian(params[:0], params, params[0:0])
            jac_img += target.jacobian_image(parameters=identities, data=jac)

        return jac_img

    def gradient(
        self,
        params: Optional[ArrayLike] = None,
        likelihood: Literal["gaussian", "poisson"] = "gaussian",
    ) -> ArrayLike:
        """Compute the gradient of the model likelihood with respect to its parameters."""

        jacobian_image = self.jacobian(params=params).flatten("data")

        data = self.target[self.window].flatten("data")
        mask = self.target[self.window].flatten("mask")
        model = self().flatten("data")
        if likelihood == "gaussian":
            weight = self.target[self.window].flatten("weight")
            gradient = backend.sum(
                jacobian_image * ((data - model) * weight * (~mask))[..., None], dim=0
            )
        elif likelihood == "poisson":
            gradient = backend.sum(
                jacobian_image * ((1 - data / model) * (~mask))[..., None],
                dim=0,
            )

        return gradient



# Metropolis-Adjusted Langevin Algorithm sampler
from typing import Optional, Sequence

import numpy as np

from .base import BaseOptimizer
from ..models import Model
from .. import config
from ..backend_obj import backend
from . import func

__all__ = ("MALA",)


class MALA(BaseOptimizer):
    """Metropolis-Adjusted Langevin Algorithm (MALA) sampler, based on:
    https://en.wikipedia.org/wiki/Metropolis-adjusted_Langevin_algorithm . This
    is a gradient-based MCMC sampler that uses the gradient of the
    log-likelihood to propose new samples. These gradient based proposals can
    lead to more efficient sampling of the parameter space. This is especially
    true when the mass_matrix is set well. A good guess for the mass matrix is
    the covariance matrix of the likelihood at the maximum likelihood point.
    Which can be found fairly easily with the LM optimizer (see the fitting
    methods tutorial).

    **Args:**
    -  `chains`: The number of MCMC chains to run in parallel. Default is 4.
    -  `epsilon`: The step size for the MALA sampler. Default is 1e-2.
    -  `mass_matrix`: The mass matrix for the MALA sampler. If None, the identity matrix is used.
    -  `progress_bar`: Whether to show a progress bar during sampling. Default is True.
    -  `likelihood`: The likelihood function to use for the MCMC sampling. Can be "gaussian" or "poisson". Default is "gaussian".
    """

    def __init__(
        self,
        model: Model,
        initial_state: Optional[Sequence] = None,
        chains=4,
        epsilon: float = 1e-2,
        mass_matrix: Optional[np.ndarray] = None,
        max_iter: int = 1000,
        progress_bar: bool = True,
        likelihood="gaussian",
        **kwargs,
    ):
        super().__init__(model, initial_state, max_iter=max_iter, **kwargs)
        self.chain = []
        if len(self.current_state.shape) == 2:
            self.chains = self.current_state.shape[0]
        else:
            self.chains = chains
        self.likelihood = likelihood
        self.epsilon = epsilon
        self.mass_matrix = mass_matrix
        self.progress_bar = progress_bar

    def density_func(self):
        """
        Returns the density of the model at the given state vector.
        This is used to calculate the likelihood of the model at the given state.
        """
        if self.likelihood == "gaussian":
            vll = backend.vmap(self.model.gaussian_log_likelihood)
        elif self.likelihood == "poisson":
            vll = backend.vmap(self.model.poisson_log_likelihood)
        else:
            raise ValueError(f"Unknown likelihood type: {self.likelihood}")

        def dens(state: np.ndarray) -> np.ndarray:
            state = backend.as_array(state, dtype=config.DTYPE, device=config.DEVICE)
            return backend.to_numpy(vll(state))

        return dens

    def density_grad_func(self):
        """
        Returns the gradient of the density of the model at the given state vector.
        This is used to calculate the gradient of the likelihood of the model at the given state.
        """
        if self.likelihood == "gaussian":
            vll_grad = backend.vmap(backend.grad(self.model.gaussian_log_likelihood))
        elif self.likelihood == "poisson":
            vll_grad = backend.vmap(backend.grad(self.model.poisson_log_likelihood))
        else:
            raise ValueError(f"Unknown likelihood type: {self.likelihood}")

        def grad(state: np.ndarray) -> np.ndarray:
            state = backend.as_array(state, dtype=config.DTYPE, device=config.DEVICE)
            return backend.to_numpy(vll_grad(state))

        return grad

    def fit(self):

        Px = self.density_func()
        dPdx = self.density_grad_func()

        initial_state = backend.to_numpy(self.current_state)
        if len(initial_state.shape) == 1:
            initial_state = np.repeat(initial_state[None, :], self.chains, axis=0)

        if self.mass_matrix is None:
            D = initial_state.shape[1]
            self.mass_matrix = np.eye(D, dtype=initial_state.dtype)

        self.chain, self.logp = func.mala(
            initial_state,
            Px,
            dPdx,
            self.max_iter,
            self.epsilon,
            self.mass_matrix,
            progress=self.progress_bar,
            desc="MALA",
        )
        # Fill model with max logp sample
        max_logp_index = np.argmax(self.logp)
        max_logp_index = np.unravel_index(max_logp_index, self.logp.shape)
        self.model.set_values(
            backend.as_array(self.chain[max_logp_index], dtype=config.DTYPE, device=config.DEVICE)
        )

        return self



import torch

from ...param import forward
from ...backend_obj import ArrayLike
from ...utils.decorators import ignore_numpy_warnings
from .._shared_methods import parametric_initialize, parametric_segment_initialize
from ...utils.parametric_profiles import exponential_np
from .. import func


def _x0_func(model_params, R, F):
    return R[4], 10 ** F[4]


class ExponentialMixin:
    """Exponential radial light profile.

    An exponential is a classical radial model used in many contexts. The
    functional form of the exponential profile is defined as:

    $$I(R) = I_e * \\exp\\left(- b_1\\left(\\frac{R}{R_e} - 1\\right)\\right)$$

    Ie is the brightness at the effective radius, and Re is the effective
    radius. $b_1$ is a constant that ensures $I_e$ is the brightness at $R_e$.

    **Parameters:**
    -    `Re`: effective radius in arcseconds
    -    `Ie`: effective surface density in flux/arcsec^2
    """

    _model_type = "exponential"
    _parameter_specs = {
        "Re": {"units": "arcsec", "valid": (0, None), "shape": (), "dynamic": True},
        "Ie": {"units": "flux/arcsec^2", "valid": (0, None), "shape": (), "dynamic": True},
    }

    @torch.no_grad()
    @ignore_numpy_warnings
    def initialize(self):
        super().initialize()

        parametric_initialize(
            self,
            self.target[self.window],
            exponential_np,
            ("Re", "Ie"),
            _x0_func,
        )

    @forward
    def radial_model(self, R: ArrayLike, Re: ArrayLike, Ie: ArrayLike) -> ArrayLike:
        return func.exponential(R, Re, Ie)


class iExponentialMixin:
    """Exponential radial light profile.

    An exponential is a classical radial model used in many contexts. The
    functional form of the exponential profile is defined as:

    $$I(R) = I_e * \\exp\\left(- b_1\\left(\\frac{R}{R_e} - 1\\right)\\right)$$

    $I_e$ is the brightness at the effective radius, and $R_e$ is the effective
    radius. $b_1$ is a constant that ensures $I_e$ is the brightness at $R_e$.

    `Re` and `Ie` are batched by their first dimension, allowing for multiple
    exponential profiles to be defined at once.

    **Parameters:**
    -    `Re`: effective radius in arcseconds
    -    `Ie`: effective surface density in flux/arcsec^2
    """

    _model_type = "exponential"
    _parameter_specs = {
        "Re": {"units": "arcsec", "valid": (0, None), "shape": (None,), "dynamic": True},
        "Ie": {"units": "flux/arcsec^2", "valid": (0, None), "shape": (None,), "dynamic": True},
    }

    @torch.no_grad()
    @ignore_numpy_warnings
    def initialize(self):
        super().initialize()

        parametric_segment_initialize(
            model=self,
            target=self.target[self.window],
            prof_func=exponential_np,
            params=("Re", "Ie"),
            x0_func=_x0_func,
            segments=self.segments,
        )

    @forward
    def iradial_model(self, i: int, R: ArrayLike, Re: ArrayLike, Ie: ArrayLike) -> ArrayLike:
        return func.exponential(R, Re[i], Ie[i])


class ExponentialPSFMixin:
    """Exponential radial light profile.

    An exponential is a classical radial model used in many contexts. The
    functional form of the exponential profile is defined as:

    $$I(R) = I_e * \\exp\\left(- b_1\\left(\\frac{R}{R_e} - 1\\right)\\right)$$

    Ie is the brightness at the effective radius, and Re is the effective
    radius. $b_1$ is a constant that ensures $I_e$ is the brightness at $R_e$.

    **Parameters:**
    -    `Re`: effective radius in pixels
    -    `Ie`: effective surface density in flux/pix^2
    """

    _model_type = "exponential"
    _parameter_specs = {
        "Re": {"units": "pix", "valid": (0, None), "shape": (), "dynamic": True},
        "Ie": {
            "units": "flux/pix^2",
            "valid": (0, None),
            "shape": (),
            "dynamic": False,
            "value": 1.0,
        },
    }

    @torch.no_grad()
    @ignore_numpy_warnings
    def initialize(self):
        super().initialize()

        parametric_initialize(
            self,
            self.target[self.window],
            exponential_np,
            ("Re", "Ie"),
            _x0_func,
        )

    @forward
    def radial_model(self, R: ArrayLike, Re: ArrayLike, Ie: ArrayLike) -> ArrayLike:
        return func.exponential(R, Re, Ie)



import astrophot as ap
import nbformat
from nbformat.v4 import new_notebook, new_markdown_cell
import pkgutil
from types import ModuleType, FunctionType
import os
from textwrap import dedent
from inspect import cleandoc, getmodule, signature

skip_methods = [
    "to_valid",
    "topological_ordering",
    "to_static",
    "to_dynamic",
    "unlink",
    "update_graph",
    "save_state",
    "load_state",
    "append_state",
    "link",
    "graphviz",
    "graph_print",
    "graph_dict",
    "from_valid",
    "fill_params",
    "fill_kwargs",
    "fill_dynamic_values",
    "clear_params",
    "build_params_list",
    "build_params_dict",
    "build_params_array",
]


def dot_path(path):
    i = path.rfind("AstroPhot")
    path = path[i + 10 :]
    path = path.replace("/", ".")
    return path[:-3]


def gather_docs(module, module_only=False):
    docs = {}
    for name in module.__all__:
        obj = getattr(module, name)
        if module_only and not isinstance(obj, ModuleType):
            continue
        if isinstance(obj, type):
            if obj.__doc__ is None:
                continue
            docs[name] = cleandoc(obj.__doc__)
            subfuncs = [docs[name]]
            for attr in dir(obj):
                if attr.startswith("_"):
                    continue
                if attr in skip_methods:
                    continue
                attrobj = getattr(obj, attr)
                if not isinstance(attrobj, FunctionType):
                    continue
                if attrobj.__doc__ is None:
                    continue
                sig = str(signature(attrobj)).replace("self,", "").replace("self", "")
                subfuncs.append(f"<u>**method:**</u> {attr}{sig}\n\n" + cleandoc(attrobj.__doc__))
            if len(subfuncs) > 1:
                docs[name] = "\n\n".join(subfuncs)
        elif isinstance(obj, FunctionType):
            if obj.__doc__ is None:
                continue
            sig = str(signature(obj))
            docs[name] = "<u>**signature:**</u> " + name + sig + "\n\n" + cleandoc(obj.__doc__)
        elif isinstance(obj, ModuleType):
            docs[name] = gather_docs(obj)
        else:
            print(f"!!!unexpected type {type(obj)}!!!")
    return docs


def make_cells(mod_dict, path, depth=2):
    print(mod_dict.keys())
    cells = []
    for k in mod_dict:
        if isinstance(mod_dict[k], str):
            cells.append(new_markdown_cell(f"{'#'*depth} {path}.{k}\n\n" + mod_dict[k]))
        elif isinstance(mod_dict[k], dict):
            print(k)
            cells += make_cells(mod_dict[k], path=path + "." + k, depth=depth + 1)
    return cells


output_dir = "docs/source/astrophotdocs"
all_ap = gather_docs(ap, True)

for submodule in all_ap:
    nb = new_notebook()
    nb.cells = [new_markdown_cell(f"# {submodule}")] + make_cells(
        all_ap[submodule], f"astrophot.{submodule}"
    )

    filename = f"{submodule}.ipynb"
    path = os.path.join(output_dir, filename)
    with open(path, "w", encoding="utf-8") as f:
        nbformat.write(nb, f)



from typing import List, Optional, Union
import numpy as np

from ..errors import SpecificationConflict, InvalidImage

from .jacobian_image import JacobianImage
from .. import config
from ..backend_obj import backend, ArrayLike
from .mixins import DataMixin
from .window import Window
from . import func

__all__ = ("PSFImage",)


class PSFImage(DataMixin):
    """Image object which represents a model of PSF (Point Spread Function).

    PSFImage inherits from the base Image class and represents the model of a point spread function.
    The point spread function characterizes the response of an imaging system to a point source or point object.

    The shape of the PSF data should be odd (for your sanity) but this is not enforced.
    """

    base_scale = 1.0

    def __init__(
        self,
        *,
        data: Optional[ArrayLike] = None,
        upsample: int = 1,
        crpix: Optional[tuple[int, int]] = None,
        filename: Optional[str] = None,
        hduext: int = 0,
        identity: str = None,
        _data: Optional[ArrayLike] = None,
        **kwargs,
    ):
        if _data is None:
            self.data = data
        else:
            self._data = _data

        super().__init__(**kwargs)
        self.upsample = upsample
        self.crpix = crpix

        if identity is None:
            self._identity = id(self)
        else:
            self._identity = identity

        if filename is not None:
            self.load(filename, hduext=hduext)
            return

    def normalize(self):
        """Normalizes the PSF image to have a sum of 1."""
        norm = backend.sum(self._data, dim=(-2, -1), keepdim=True)
        self._data = self._data / norm
        self._weight = self._weight * norm**2

    @property
    def identity(self):
        return self._identity

    @property
    def window(self) -> Window:
        return Window(window=((0, 0), self._data.shape[:2]), image=self)

    @property
    def data(self):
        """The image data, which is a tensor of pixel values."""
        return backend.transpose(self._data, 1, 0)

    @data.setter
    def data(self, value: Optional[ArrayLike]):
        """Set the image data. If value is None, the data is initialized to an empty tensor."""
        if value is None:
            self._data = backend.ones((1, 1), dtype=config.DTYPE, device=config.DEVICE)
        else:
            assert all(
                s % 2 == 1 for s in value.shape[-2:]
            ), "PSF data shape must be odd in both dimensions."
            # Transpose since pytorch uses (j, i) indexing when (i, j) is more natural for coordinates
            self._data = backend.transpose(
                backend.as_array(value, dtype=config.DTYPE, device=config.DEVICE), 1, 0
            )

    @property
    def upsample(self) -> int:
        return self._upsample

    @upsample.setter
    def upsample(self, value: int):
        if value < 1:
            raise ValueError("upsample factor must be a positive integer.")
        self._upsample = int(value)

    @property
    def pad(self) -> int:
        return max(self._data.shape[-2:]) // 2

    @property
    def crpix(self):
        if self._crpix is None:
            return np.array([self._data.shape[-2] // 2, self._data.shape[-1] // 2])
        return self._crpix

    @crpix.setter
    def crpix(self, value):
        if value is None:
            self._crpix = None
        else:
            self._crpix = np.array(value)

    @property
    def pixel_area(self):
        return backend.as_array(1.0 / self.upsample**2, dtype=config.DTYPE, device=config.DEVICE)

    @property
    def pixelscale(self):
        return backend.as_array(1.0 / self.upsample, dtype=config.DTYPE, device=config.DEVICE)

    @property
    def flip_ra_axis(self):
        return False

    @property
    def zeropoint(self):
        return None

    def flatten(self, attribute: str = "data") -> ArrayLike:
        return backend.flatten(getattr(self, attribute), end_dim=1)

    def pixel_collecting_area(self, *args, **kwargs):
        return 1.0

    def targpixel_to_mypixel(self, I_, J_):
        """
        Convert between coordinate spaces. "targpixel" refers to the pixel
        coordinates of the target of this PSF, which have the origin at the
        center of the PSF and a step of 1 corresponds to one target pixel
        length. "mypixel" refers to the pixel coordinates of this PSF image,
        which have an origin at the center of the [0,0] pixel and a step of 1
        corresponds to one PSF pixel length (which is 1/upsample of a target
        pixel length).
        """
        return I_ * self.upsample + self.crpix[0], J_ * self.upsample + self.crpix[1]

    def mypixel_to_targpixel(self, i, j):
        """
        Convert between coordinate spaces. "targpixel" refers to the pixel
        coordinates of the target of this PSF, which have the origin at the
        center of the PSF and a step of 1 corresponds to one target pixel
        length. "mypixel" refers to the pixel coordinates of this PSF image,
        which have an origin at the center of the [0,0] pixel and a step of 1
        corresponds to one PSF pixel length (which is 1/upsample of a target
        pixel length).
        """
        return (i - self.crpix[0]) / self.upsample, (j - self.crpix[1]) / self.upsample

    def pixel_center_meshgrid(self, window=None, pad=0, upsample=1) -> tuple[ArrayLike, ArrayLike]:
        """Get a meshgrid of pixel coordinates in the image, centered on the pixel grid."""
        if window is None:
            window = self.window
        return func.pixel_center_meshgrid(window.extent, pad, upsample, config.DTYPE, config.DEVICE)

    def pixel_corner_meshgrid(self, window=None, pad=0, upsample=1) -> tuple[ArrayLike, ArrayLike]:
        """Get a meshgrid of pixel coordinates in the image, with corners at the pixel grid."""
        if window is None:
            window = self.window
        return func.pixel_corner_meshgrid(window.extent, pad, upsample, config.DTYPE, config.DEVICE)

    def pixel_simpsons_meshgrid(
        self, window=None, pad=0, upsample=1
    ) -> tuple[ArrayLike, ArrayLike]:
        """Get a meshgrid of pixel coordinates in the image, with Simpson's rule sampling."""
        if window is None:
            window = self.window
        return func.pixel_simpsons_meshgrid(
            window.extent, pad, upsample, config.DTYPE, config.DEVICE
        )

    def pixel_quad_meshgrid(
        self, window=None, pad=0, upsample=1, order=3
    ) -> tuple[ArrayLike, ArrayLike]:
        """Get a meshgrid of pixel coordinates in the image, with quadrature sampling."""
        if window is None:
            window = self.window
        return func.pixel_quad_meshgrid(
            window.extent, pad, upsample, config.DTYPE, config.DEVICE, order=order
        )

    def coordinate_center_meshgrid(self) -> tuple[ArrayLike, ArrayLike]:
        i, j = self.pixel_center_meshgrid()
        return self.mypixel_to_targpixel(i, j)

    def coordinate_corner_meshgrid(self) -> tuple[ArrayLike, ArrayLike]:
        i, j = self.pixel_corner_meshgrid()
        return self.mypixel_to_targpixel(i, j)

    def coordinate_simpsons_meshgrid(self) -> tuple[ArrayLike, ArrayLike]:
        i, j = self.pixel_simpsons_meshgrid()
        return self.mypixel_to_targpixel(i, j)

    def coordinate_quad_meshgrid(self, order=3) -> tuple[ArrayLike, ArrayLike, ArrayLike]:
        i, j, _ = self.pixel_quad_meshgrid(order=order)
        return self.mypixel_to_targpixel(i, j)

    def get_indices(self, other: Window):
        if other.image is self:
            return slice(max(0, other.i_low), min(self._data.shape[0], other.i_high)), slice(
                max(0, other.j_low), min(self._data.shape[1], other.j_high)
            )
        if other.image.identity == self.identity:
            shift = np.round(self.crpix - other.crpix).astype(int)
            return slice(
                min(max(0, other.i_low + shift[0]), self._data.shape[0]),
                max(0, min(other.i_high + shift[0], self._data.shape[0])),
            ), slice(
                min(max(0, other.j_low + shift[1]), self._data.shape[1]),
                max(0, min(other.j_high + shift[1], self._data.shape[1])),
            )
        raise RuntimeError(
            f"Cannot get indices for window with different image! Window image: {other.image.name}, self image: {self.name}"
        )

    def jacobian_image(
        self,
        parameters: Optional[List[str]] = None,
        data: Optional[ArrayLike] = None,
        **kwargs,
    ) -> JacobianImage:
        """
        Construct a blank `JacobianImage` object formatted like this current `PSFImage` object. Mostly used internally.
        """
        if parameters is None:
            data = None
            parameters = []
        elif data is None:
            data = backend.zeros(
                (*self._data.shape, len(parameters)),
                dtype=config.DTYPE,
                device=config.DEVICE,
            )
        kwargs = {
            "CD": ((1, 0), (0, 1)),
            "crpix": self.crpix,
            "crtan": (0, 0),
            "crval": (0, 0),
            "identity": self.identity,
            "_data": data,
            **kwargs,
        }
        return JacobianImage(parameters=parameters, **kwargs)

    def model_image(self, window=None, **kwargs) -> "PSFImage":
        """
        Construct a blank `ModelImage` object formatted like this current `TargetImage` object. Mostly used internally.
        """
        if window is None:
            window = self.window
        si, sj = self.get_indices(window)
        kwargs = {
            "_data": backend.zeros_like(self._data[si, sj]),
            "crpix": self.crpix - np.array((si.start, sj.start)),
            "upsample": self.upsample,
            "identity": self.identity,
            **kwargs,
        }
        return PSFImage(**kwargs)

    def fits_info(self) -> dict:
        return {
            "UPSMPL": self.upsample,
            "CRPIX1": self.crpix[0],
            "CRPIX2": self.crpix[1],
            "IDNTY": self.identity,
        }

    def __iadd__(self, other: Union["PSFImage", ArrayLike]):
        if isinstance(other, (int, float, backend.array_type)):
            self._data = self._data + other
            return self
        if not isinstance(other, PSFImage):
            raise InvalidImage(f"PSF images can only add with each other, not: {type(other)}")

        if self.upsample != other.upsample:
            raise SpecificationConflict("Cannot add PSF images with different upsample factors.")

        islice = slice(
            max(0, self.crpix[0] - other.crpix[0]),
            min(self._data.shape[-2], self.crpix[0] + other._data.shape[-2] - other.crpix[0]),
        )
        jslice = slice(
            max(0, self.crpix[1] - other.crpix[1]),
            min(self._data.shape[-1], self.crpix[1] + other._data.shape[-1] - other.crpix[1]),
        )
        self._data = backend.add_at_indices(self._data, (islice, jslice), other._data)
        return self

    def __sub__(self, other: "PSFImage"):
        return self.copy(_data=self._data - other._data)

    def copy_kwargs(self, **kwargs) -> dict:
        kwargs = {
            "_data": backend.copy(self._data),
            "crpix": self.crpix,
            "upsample": self.upsample,
            "identity": self.identity,
            **kwargs,
        }
        return super().copy_kwargs(**kwargs)

    def copy(self, **kwargs):
        """Produce a copy of this image with all of the same properties. This
        can be used when one wishes to make temporary modifications to
        an image and then will want the original again.

        """
        return self.__class__(**self.copy_kwargs(**kwargs))

    def reduce(self, scale: int):
        return self.copy(upsample=self.upsample * scale)

    def get_window(self, other: Union[Window, "Image"], indices=None, **kwargs):
        """Get a new image object which is a window of this image
        corresponding to the other image's window. This will return a
        new image object with the same properties as this one, but with
        the data cropped to the other image's window.

        """
        if indices is None:
            indices = self.get_indices(other if isinstance(other, Window) else other.window)
        new_img = self.copy(
            _data=self._data[indices],
            crpix=self.crpix - np.array((indices[0].start, indices[1].start)),
            **kwargs,
        )
        return new_img

    def __getitem__(self, *args):
        if len(args) == 1 and isinstance(args[0], Window):
            return self.get_window(args[0])
        return super().__getitem__(*args)



from functools import lru_cache
from scipy.special import binom
import numpy as np


@lru_cache(maxsize=1024)
def coefficients(n: int, m: int) -> list[tuple[int, float]]:
    C = []
    for k in range(int((n - abs(m)) / 2) + 1):
        C.append(
            (
                k,
                (-1) ** k * binom(n - k, k) * binom(n - 2 * k, (n - abs(m)) / 2 - k),
            )
        )
    return C


def zernike_n_m_list(n: int) -> list[tuple[int, int]]:
    nm = []
    for n_i in range(n + 1):
        for m_i in range(-n_i, n_i + 1, 2):
            nm.append((n_i, m_i))
    return nm


def zernike_n_m_modes(rho: np.ndarray, phi: np.ndarray, n: int, m: int) -> np.ndarray:
    Z = np.zeros_like(rho)
    for k, c in coefficients(n, m):
        R = rho ** (n - 2 * k)
        T = 1.0
        if m < 0:
            T = np.sin(abs(m) * phi)
        elif m > 0:
            T = np.cos(m * phi)

        Z = Z + c * R * T
    return Z * (rho <= 1).astype(np.float64)


def zernike_basis(order: int, N: int) -> np.ndarray:
    nm = zernike_n_m_list(order)
    X, Y = np.meshgrid(
        np.linspace(-1, 1, N) * (N - 1) / N,
        np.linspace(-1, 1, N) * (N - 1) / N,
        indexing="ij",
    )
    R = np.sqrt(X**2 + Y**2)
    Phi = np.arctan2(Y, X)
    basis = []
    for n, m in nm:
        basis.append(zernike_n_m_modes(R, Phi, n, m))
    return np.stack(basis, axis=0)



from typing import Optional
from ..backend_obj import backend
from ..param import forward
from ..errors import SpecificationConflict
from .base import Model
from .mixins import SampleMixin, GradMixin
from .model_object import ComponentModel
from ..image import TargetImage, Window, TargetImageBatch, WindowBatch
from . import func
from .. import config


class BatchModel(GradMixin, SampleMixin, Model):
    """A batch of models that all share the same window/target.

    This can for example be used to model a crowded area of the sky with many
    overlapping sources, or to model a single object that is represented by many
    components (consider this a generalization of the Multi-gaussian expansion
    model). If you want to model the same object in multiple images, see the
    BatchSceneModel instead.

    **Note:** any model parameters that you wish to batch over must be set to
        dynamic=True. See [caskade hierarchical
        models](https://caskade.readthedocs.io/en/latest/notebooks/HierarchicalModels.html)
        for more details.
    """

    usable = True
    _model_type = "batch"

    def __init__(self, *, model: ComponentModel = None, **kwargs):
        super().__init__(**kwargs)
        assert isinstance(
            model, ComponentModel
        ), "BatchModel must be initialized with a ComponentModel instance."
        self.hierarchical_link("model", model)

    def initialize(self):
        self.model.initialize()

    @property
    def target(self) -> Optional[TargetImage]:
        return self.model.target

    @target.setter
    def target(self, target: Optional[TargetImage]):
        pass

    @property
    def window(self) -> Optional[Window]:
        """The window defines a region on the sky in which this model will be
        optimized and evaluated. Two models with non-overlapping windows are in
        effect independent of each other. If there is another model with a
        window that spans both of them, then they are tenuously connected.

        If not provided, the model will assume a window equal to the target it
        is fitting. Note that in this case the window is not explicitly set to
        the target window, so if the model is moved to another target then the
        fitting window will also change.

        """
        return self.model.window

    @window.setter
    def window(self, window):
        pass

    @property
    def mask(self):
        return self.model.mask

    @mask.setter
    def mask(self, mask):
        pass

    @forward
    def __call__(self, model_params=None, model_dims=None, **kwargs):
        psf, upsample, pad = self.model._prep_psf()
        working_image = self.target.model_image(self.window)
        I, J = self.model._pixel_meshgridder(self.target, self.window, pad, upsample)
        Z = backend.vmap(
            self.model.sample,
            in_dims=(None, None, None, None, None, model_dims),
        )(
            I,
            J,
            None,
            pad,
            upsample,
            model_params,
        )
        Z = backend.sum(Z, dim=0)
        if psf is not None and not self.model.internal_psf:
            if isinstance(psf, Model):
                psf = psf()._data
            if psf.shape != (1, 1):  # skip if identity PSF
                Z = func.convolve(Z, psf)
                Z = Z[pad : Z.shape[0] - pad, pad : Z.shape[1] - pad]
                Z = func.downsample(Z, upsample)
        working_image._data = Z
        return working_image


class BatchSceneModel(GradMixin, Model):
    """A single model as viewed in multiple images.

    This model is quite restrictive in its use, but can provide a significant
    speedup by vectorizing the evaluation of the model. Some key things to keep
    in mind:

    - All model parameters that you wish to batch over must be set to
      dynamic=True. See [caskade hierarchical
      models](https://caskade.readthedocs.io/en/latest/notebooks/HierarchicalModels.html)
      for more details.

    - You must use a TargetImageBatch as the target, meaning that all the images
      must be the same number of pixels.

    - You must use a WindowBatch as the window (or none to just use the full
      images), meaning that the windows must all be the same shape in pixels.

    - The model you provide must have a TargetImage target and Window window,
      and these must have the same number of pixels as the batched versions.

    - If the base model has a PSFModel as its PSF then you cannot override it
      (that's part of the model), otherwise the PSF from the TargetImageBatch
      will override the model PSF for all images.
    """

    usable = True
    _model_type = "batch scene"

    def __init__(self, *, model: Model = None, **kwargs):
        super().__init__(**kwargs)
        if not isinstance(model.target, TargetImage):
            raise SpecificationConflict(
                f"BatchSceneModel can only be used with models that have a TargetImage as their target, not a {type(model.target).__name__}."
            )
        self.hierarchical_link("model", model)

    def initialize(self):
        self.model.initialize()

    @property
    def target(self) -> Optional[TargetImageBatch]:
        return self._target

    @target.setter
    def target(self, target: TargetImageBatch):
        assert isinstance(
            target, TargetImageBatch
        ), "BatchSceneModel target must be a TargetImageBatch."
        self._target = target

    @property
    def window(self) -> WindowBatch:
        if self._window is None:
            return self._target.window
        return self._window

    @window.setter
    def window(self, window):
        assert window is None or isinstance(
            window, WindowBatch
        ), "BatchSceneModel window must be a WindowBatch."
        self._window = window

    @forward
    def __call__(self, model_params=None, model_dims=None, **kwargs):
        working_image = self.target.model_image(self.window)
        crtan = self.target.crtan
        shift = backend.as_array(
            self.window.origin_shifter(self.model.window), dtype=config.DTYPE, device=config.DEVICE
        )
        crpix = self.target.crpix + shift
        CD = self.target.CD
        psf = self.target.psf_stack
        psf_batch = None if psf is None else 0
        working_image._data = backend.vmap(
            lambda *args: self.model(*args)._data, in_dims=(0, 0, 0, psf_batch, model_dims)
        )(CD, crtan, crpix, psf, model_params)
        return working_image



from ...backend_obj import backend, ArrayLike

C1 = 4 / 405
C2 = 46 / 25515
C3 = 131 / 1148175
C4 = -2194697 / 30690717750


def sersic_n_to_b(n: float) -> float:
    """Compute the `b(n)` for a sersic model. This factor ensures that
    the $R_e$ and $I_e$ parameters do in fact correspond
    to the half light values and not some other scale
    radius/intensity.

    """
    x = 1 / n
    return 2 * n - 1 / 3 + x * (C1 + x * (C2 + x * (C3 + C4 * x)))


def sersic(R: ArrayLike, n: ArrayLike, Re: ArrayLike, Ie: ArrayLike) -> ArrayLike:
    """Seric 1d profile function, specifically designed for pytorch
    operations

    **Args:**
    -  `R`: Radii tensor at which to evaluate the sersic function
    -  `n`: sersic index restricted to n > 0.36
    -  `Re`: Effective radius in the same units as R
    -  `Ie`: Effective surface density
    """
    bn = sersic_n_to_b(n)
    return Ie * backend.exp(-bn * ((R / Re) ** (1 / n) - 1))



from typing import Optional, Tuple
import torch
import numpy as np

from .model_object import ComponentModel
from ..utils.decorators import ignore_numpy_warnings, combine_docstrings
from . import func
from .. import config
from ..backend_obj import backend, ArrayLike
from ..param import forward

__all__ = ["MultiGaussianExpansion"]


@combine_docstrings
class MultiGaussianExpansion(ComponentModel):
    """Model that represents a galaxy as a sum of multiple Gaussian
    profiles. The model is defined as:

    $$I(R) = \\sum_i {\\rm flux}_i * \\exp(-0.5*(R_i / \\sigma_i)^2) / (2 * \\pi * q_i * \\sigma_i^2)$$

    where $R_i$ is a radius computed using $q_i$ and $PA_i$ for that component. All components share the same center.

    **Parameters:**
    -    `q`: axis ratio to scale minor axis from the ratio of the minor/major axis b/a, this parameter is unitless, it is restricted to the range (0,1)
    -    `PA`: position angle of the semi-major axis relative to the image positive x-axis in radians, it is a cyclic parameter in the range [0,pi)
    -    `sigma`: standard deviation of each Gaussian
    -    `flux`: amplitude of each Gaussian
    """

    _model_type = "mge"
    _parameter_specs = {
        "q": {"units": "b/a", "valid": (0, 1), "shape": (None,), "dynamic": True},
        "PA": {
            "units": "radians",
            "valid": (0, np.pi),
            "cyclic": True,
            "dynamic": True,
        },  # No shape for PA since there are two options, use with caution
        "sigma": {"units": "arcsec", "valid": (0, None), "shape": (None,), "dynamic": True},
        "flux": {"units": "flux", "shape": (None,), "dynamic": True},
    }
    usable = True

    def __init__(self, *args, n_components: Optional[int] = None, **kwargs):
        super().__init__(*args, **kwargs)
        if n_components is None:
            for key in ("q", "sigma", "flux"):
                if self[key].value is not None:
                    self.n_components = self[key].value.shape[0]
                    break
            else:
                self.n_components = 1
        else:
            self.n_components = int(n_components)

    @torch.no_grad()
    @ignore_numpy_warnings
    def initialize(self):
        super().initialize()

        target_area = self.target[self.window]
        dat = backend.to_numpy(target_area._data).copy()
        mask = backend.to_numpy(target_area._mask)
        dat[mask] = np.median(dat[~mask])

        edge = np.concatenate((dat[:, 0], dat[:, -1], dat[0, :], dat[-1, :]))
        edge_average = np.nanmedian(edge)
        dat -= edge_average

        if not self.sigma.initialized:
            self.sigma.value = np.logspace(
                np.log10(target_area.pixelscale.item() * 3),
                max(target_area.data.shape) * target_area.pixelscale.item() * 0.7,
                self.n_components,
            )
        if not self.flux.initialized:
            self.flux.value = (np.sum(dat) / self.n_components) * np.ones(self.n_components)

        if self.PA.initialized and self.q.initialized:
            return

        x, y = target_area.coordinate_center_meshgrid()
        x = backend.to_numpy(x - self.center.value[0])
        y = backend.to_numpy(y - self.center.value[1])
        mu20 = np.median(dat * np.abs(x))
        mu02 = np.median(dat * np.abs(y))
        mu11 = np.median(dat * x * y / np.sqrt(np.abs(x * y) + self.softening**2))
        # mu20 = np.median(dat * x**2)
        # mu02 = np.median(dat * y**2)
        # mu11 = np.median(dat * x * y)
        M = np.array([[mu20, mu11], [mu11, mu02]])
        ones = np.ones(self.n_components)
        if not self.PA.initialized:
            if np.any(np.iscomplex(M)) or np.any(~np.isfinite(M)):
                self.PA.value = ones * np.pi / 2
            else:
                self.PA.value = ones * (0.5 * np.arctan2(2 * mu11, mu20 - mu02) - np.pi / 2) % np.pi
        if not self.q.initialized:
            l = np.sort(np.linalg.eigvals(M))
            if np.any(np.iscomplex(l)) or np.any(~np.isfinite(l)):
                l = (0.7, 1.0)
            self.q.value = ones * np.clip(np.sqrt(l[0] / l[1]), 0.1, 0.9)

    @forward
    def transform_coordinates(
        self, x: ArrayLike, y: ArrayLike, q: ArrayLike, PA: ArrayLike
    ) -> Tuple[ArrayLike, ArrayLike]:
        x, y = super().transform_coordinates(x, y)
        if np.prod(PA.shape) == 1:
            x, y = func.rotate(-(PA + np.pi / 2), x, y)
            x = x * backend.ones(
                (q.shape[0], *[1] * x.ndim), dtype=config.DTYPE, device=config.DEVICE
            )
            y = y * backend.ones(
                (q.shape[0], *[1] * y.ndim), dtype=config.DTYPE, device=config.DEVICE
            )
        else:
            x, y = backend.vmap(lambda pa: func.rotate(-(pa + np.pi / 2), x, y))(PA)
        y = backend.vmap(lambda q, y: y / q)(q, y)
        return x, y

    @forward
    def brightness(
        self, x: ArrayLike, y: ArrayLike, flux: ArrayLike, sigma: ArrayLike, q: ArrayLike
    ) -> ArrayLike:
        x, y = self.transform_coordinates(x, y)
        R = self.radius_metric(x, y)
        return backend.sum(
            backend.vmap(
                lambda A, r, sig, _q: (A / (2 * np.pi * _q * sig**2))
                * backend.exp(-0.5 * (r / sig) ** 2)
            )(flux, R, sigma, q),
            dim=0,
        )



import torch
from ...backend_obj import backend, ArrayLike
import numpy as np

sq_2pi = np.sqrt(2 * np.pi)


def gaussian(R: ArrayLike, sigma: ArrayLike, flux: ArrayLike) -> ArrayLike:
    """Gaussian 1d profile function, specifically designed for pytorch
    operations.

    **Args:**
    -  `R`: Radii tensor at which to evaluate the gaussian function
    -  `sigma`: Standard deviation of the gaussian in the same units as R
    -  `flux`: Central surface density
    """
    return (flux / (sq_2pi * sigma)) * backend.exp(-0.5 * (R / sigma) ** 2)



from typing import Optional

import torch
import numpy as np

from .base import Model
from .model_object import ComponentModel
from ..image import ModelImage
from ..utils.decorators import ignore_numpy_warnings, combine_docstrings
from ..utils.interpolate import interp2d
from ..image import Window, PSFImage
from ..errors import SpecificationConflict
from ..param import forward
from ..backend_obj import backend, ArrayLike
from .. import config
from . import func

__all__ = ("PointSource",)


@combine_docstrings
class PointSource(ComponentModel):
    """Describes a point source in the image, this is a delta function at
    some position in the sky. This is typically used to describe
    stars, supernovae, very small galaxies, quasars, asteroids or any
    other object which can essentially be entirely described by a
    position and total flux (no structure).

    **Parameters:**
    -    `flux`: The total flux of the point source

    """

    _model_type = "point"
    _parameter_specs = {
        "flux": {"units": "flux", "valid": (0, None), "shape": (), "dynamic": True},
    }
    internal_psf = True
    usable = True

    def __init__(self, *args, integrate_mode="none", **kwargs):
        super().__init__(*args, integrate_mode=integrate_mode, **kwargs)

    @torch.no_grad()
    @ignore_numpy_warnings
    def initialize(self):
        super().initialize()
        if self.psf is None:
            raise SpecificationConflict("PointSource needs a psf!")

        if self.flux.initialized:
            return
        target_area = self.target[self.window]
        dat = backend.to_numpy(target_area._data).copy()
        mask = backend.to_numpy(target_area._mask)
        dat[mask] = np.median(dat[~mask])

        edge = np.concatenate((dat[:, 0], dat[:, -1], dat[0, :], dat[-1, :]))
        edge_average = np.median(edge)
        self.flux.value = np.abs(np.sum(dat - edge_average))

    @property
    def integrate_mode(self):
        return "none"

    @integrate_mode.setter
    def integrate_mode(self, value):
        if value != "none":
            config.logger.warning(
                "PointSource models are restricted to integrate mode of 'none', ignoring integrate_mode setting."
            )

    # Psf convolution should be on by default since this is a delta function
    @property
    def psf_convolve(self):
        return True

    @psf_convolve.setter
    def psf_convolve(self, value):
        pass

    def _prep_psf(self):
        psf = self.psf

        if isinstance(psf, PSFImage):
            return psf._data, psf.upsample, 0
        if isinstance(psf, Model):
            return psf, psf.upsample, 0
        return None, 1, 0

    @forward
    def sample(
        self,
        I_: ArrayLike,
        J_: ArrayLike,
        psf: ArrayLike = None,
        crop: int = 0,
        downsample: int = 1,
        center=None,
        flux=None,
        _CD=None,
        _crtan=None,
        _crpix=None,
    ):
        if isinstance(psf, Model):
            psf = psf()._data
        if _CD is None:
            i0, j0 = self.target.plane_to_pixel(*center)
        else:
            i0, j0 = self.target.plane_to_pixel(*center, CD=_CD, crtan=_crtan, _crpix=_crpix)
        Z = interp2d(
            psf,
            (I_ - i0) * downsample + (psf.shape[0] // 2),
            (J_ - j0) * downsample + (psf.shape[1] // 2),
        )
        Z = self._pixel_integrator(Z)
        Z = Z * flux
        Z = func.downsample(Z, downsample)
        return Z



from .mixins import (
    SersicMixin,
    iSersicMixin,
    ExponentialMixin,
    iExponentialMixin,
    GaussianMixin,
    iGaussianMixin,
    FerrerMixin,
    iFerrerMixin,
    KingMixin,
    iKingMixin,
    MoffatMixin,
    iMoffatMixin,
    NukerMixin,
    iNukerMixin,
    SplineMixin,
    iSplineMixin,
    RadialMixin,
    WedgeMixin,
    RayMixin,
    SuperEllipseMixin,
    FourierEllipseMixin,
    WarpMixin,
    TruncationMixin,
)
from .galaxy_model_object import GalaxyModel

radial_models = (
    SersicMixin,
    ExponentialMixin,
    GaussianMixin,
    FerrerMixin,
    KingMixin,
    MoffatMixin,
    NukerMixin,
    SplineMixin,
)

__all__ = []
for mixin in radial_models:
    # Galaxy Model
    g_mixin = type(
        mixin.__name__[:-5] + "Galaxy", (mixin, RadialMixin, GalaxyModel), {"usable": True}
    )
    globals()[g_mixin.__name__] = g_mixin
    __all__.append(g_mixin.__name__)

    # Truncated Galaxy Model
    t_mixin = type(
        "T" + mixin.__name__[:-5] + "Galaxy",
        (TruncationMixin, mixin, RadialMixin, GalaxyModel),
        {"usable": True},
    )
    globals()[t_mixin.__name__] = t_mixin
    __all__.append(t_mixin.__name__)

    for n, p in zip(
        ("SuperEllipse", "FourierEllipse", "Warp"),
        (SuperEllipseMixin, FourierEllipseMixin, WarpMixin),
    ):
        # Galaxy Model with additional perturbation mixin
        g_mixin = type(
            mixin.__name__[:-5] + n, (mixin, RadialMixin, p, GalaxyModel), {"usable": True}
        )
        globals()[g_mixin.__name__] = g_mixin
        __all__.append(g_mixin.__name__)

        # Truncated Galaxy Model with additional perturbation mixin
        t_mixin = type(
            "T" + mixin.__name__[:-5] + n,
            (TruncationMixin, mixin, RadialMixin, p, GalaxyModel),
            {"usable": True},
        )
        globals()[t_mixin.__name__] = t_mixin
        __all__.append(t_mixin.__name__)

iradial_models = (
    iSersicMixin,
    iExponentialMixin,
    iGaussianMixin,
    iFerrerMixin,
    iKingMixin,
    iMoffatMixin,
    iNukerMixin,
    iSplineMixin,
)

for mixin in iradial_models:
    # Ray Galaxy Model
    r_mixin = type(mixin.__name__[1:-5] + "Ray", (mixin, RayMixin, GalaxyModel), {"usable": True})
    globals()[r_mixin.__name__] = r_mixin
    __all__.append(r_mixin.__name__)

    # Wedge Galaxy Model
    w_mixin = type(
        mixin.__name__[1:-5] + "Wedge", (mixin, WedgeMixin, GalaxyModel), {"usable": True}
    )
    globals()[w_mixin.__name__] = w_mixin
    __all__.append(w_mixin.__name__)



import pytest

import astrophot as ap
from astrophot.param import Param

from utils import make_basic_sersic


def test_param():

    a = Param("a", value=1.0, uncertainty=0.1, valid=(0, 2), prof=1.0)
    assert isinstance(a.uncertainty, ap.backend.array_type), "uncertainty should be a tensor"
    assert isinstance(a.prof, ap.backend.array_type), "prof should be a tensor"
    assert a.initialized, "parameter should be marked as initialized"
    assert a.soft_valid(a.value) == a.value, "soft valid should return the value if not near limits"
    assert (
        a.soft_valid(-1 * ap.backend.ones_like(a.value)) > a.valid[0]
    ), "soft valid should push values inside the limits"
    assert (
        a.soft_valid(3 * ap.backend.ones_like(a.value)) < a.valid[1]
    ), "soft valid should push values inside the limits"

    b = Param("b", value=[2.0, 3.0], uncertainty=[0.1, 0.1], valid=(1, None))
    assert ap.backend.all(
        b.soft_valid(-1 * ap.backend.ones_like(b.value)) > b.valid[0]
    ), "soft valid should push values inside the limits"
    assert b.prof is None

    c = Param("c", value=lambda P: P.a.value, valid=(None, 4.0))
    c.link(a)
    assert c.initialized, "pointer should be marked as initialized"
    assert c.uncertainty is None


def test_module():

    target = make_basic_sersic()
    model1 = ap.Model(name="test_model_1", model_type="sersic galaxy model", target=target)
    model2 = ap.Model(name="test_model_2", model_type="sersic galaxy model", target=target)
    model = ap.Model(name="test", model_type="group model", target=target, models=[model1, model2])
    model.initialize()

    U = ap.backend.ones_like(model.get_values()) * 0.1
    model.set_values(U, attribute="uncertainty")

    paramsu = model.get_values(attribute="uncertainty")
    assert ap.backend.all(ap.backend.isfinite(paramsu)), "All parameters should be finite"

    paramsn = model.build_params_array_names()
    assert all(isinstance(name, str) for name in paramsn), "All parameter names should be strings"

    paramsun = model.build_params_array_units()
    assert all(isinstance(unit, str) for unit in paramsun), "All parameter units should be strings"



from .base import AstroPhotError

__all__ = ("InvalidTarget", "UnrecognizedModel")


class InvalidTarget(AstroPhotError):
    """
    Catches when a target object is assigned incorrectly.
    """


class UnrecognizedModel(AstroPhotError):
    """
    Raised when the user tries to invoke a model that does not exist.
    """



import astrophot as ap
import numpy as np
from utils import make_basic_gaussian_psf
import pytest

# torch.autograd.set_detect_anomaly(True)
######################################################################
# PSF Model Objects
######################################################################


@pytest.mark.parametrize("model_type", ap.models.PSFModel.List_Models(usable=True, types=True))
def test_all_psfmodel_sample(model_type):
    if model_type == "airy psf model" and ap.backend.backend == "jax":
        pytest.skip(
            "Skipping airy psf model, JAX does not support bessel_j1 with finite derivatives it seems"
        )
    if any(t in model_type for t in ["warp", "fourier"]):
        pytest.skip("Skipping warp and fourier psf models, which are slow")

    target = make_basic_gaussian_psf(pixelscale=0.8)
    MODEL = ap.Model(
        name="test_model",
        model_type=model_type,
        target=target,
        normalize_psf=False,
    )
    for p in MODEL.all_params:
        if p.units in ["flux", "flux/pix^2"]:
            p.to_dynamic(None)
    MODEL.initialize()
    for p in MODEL.all_params:
        if p.units in ["flux", "flux/pix^2"]:
            p.to_dynamic(p.value * 1.5)
        if p.units == "pix" and not p.name == "center":
            p.to_dynamic(p.value + 0.5)
    print(MODEL)
    for P in MODEL.dynamic_params:
        assert P.value is not None, (
            f"Model type {model_type} parameter {P} should not be None after initialization",
        )
    img = MODEL()

    assert ap.backend.all(
        ap.backend.isfinite(img.data)
    ), "Model should evaluate a real number for the full image"

    if model_type == "pixelated psf model":
        psf = ap.utils.initialize.gaussian_psf(3 * 0.8, 25, 0.8)
        MODEL.pixels.value = psf / np.sum(psf)

    assert ap.backend.all(
        ap.backend.isfinite(MODEL.jacobian().data)
    ), "Model should evaluate a real number for the jacobian"

    res = ap.fit.LM(MODEL, max_iter=5).fit()

    assert len(res.loss_history) >= 2, "Optimizer must be able to find steps to improve the model"

    if res.message == "success":
        # Be less strict if fit succeeded quickly
        assert res.loss_history[-1] < res.loss_history[0], (
            f"Model {model_type} should fit to the target image, but did not. "
            f"Initial loss: {res.loss_history[0]}, Final loss: {res.loss_history[-1]}"
        )
    else:
        assert ((res.loss_history[0] - 1) > (2 * (res.loss_history[-1] - 1))) or (
            res.loss_history[-1] < 1.0
        ), (
            f"Model {model_type} should fit to the target image, but did not. "
            f"Initial loss: {res.loss_history[0]}, Final loss: {res.loss_history[-1]}"
        )



from typing import Optional

import numpy as np

from ..param import forward
from .base import Model
from . import func
from ..image import TargetImage, ModelImage, PSFImage
from ..utils.initialize import recursive_center_of_mass
from ..utils.decorators import ignore_numpy_warnings, combine_docstrings
from .. import config
from ..backend_obj import backend, ArrayLike
from ..errors import InvalidTarget
from .mixins import SampleMixin, GradMixin

__all__ = ("ComponentModel",)


@combine_docstrings
class ComponentModel(GradMixin, SampleMixin, Model):
    """Component of a model for an object in an image.

    This is a single component of an image model. It has a position on the sky
    determined by `center` and may or may not be convolved with a PSF to represent some data.

    **Parameters:**
    -  `center`: The center of the component in arcseconds [x, y] defined on the tangent plane.

    **Options:**
    -  `psf_convolve`: Whether to convolve the model with a PSF. (bool)

    """

    _parameter_specs = {"center": {"units": "arcsec", "shape": (2,), "dynamic": True}}

    usable = False
    psf_convolve = True
    internal_psf = False

    _options = ("psf_convolve",)

    def __init__(self, *args, psf=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.psf = psf
        self.saveattrs.add("window.extent")

    @property
    def target(self):
        return self._target

    @target.setter
    def target(self, tar):
        if tar is None:
            self._target = None
            return
        elif not isinstance(tar, TargetImage):
            raise InvalidTarget(
                f"AstroPhot {self.__class__.__name__} target must be a TargetImage instance."
            )
        try:
            del self._target  # Remove old target if it exists
        except AttributeError:
            pass
        self._target = tar

    @property
    def psf(self):
        if self._psf is None:
            return self.target.psf
        return self._psf

    @psf.setter
    def psf(self, psf):
        try:
            del self._psf  # Remove old psf if it exists
        except AttributeError:
            pass
        if psf is None:
            self._psf = None
        elif isinstance(psf, PSFImage):
            self._psf = psf
        elif isinstance(psf, Model):
            self._psf = psf
        else:
            self._psf = PSFImage(data=psf)
            config.logger.warning(
                f"PSF provided to {self.__class__.__name__} was not a PSFImage or Model instance, so it was converted to a PSFImage assuming no upsampling."
            )

    def _prep_psf(self):
        if not self.psf_convolve:
            return None, 1, 0
        psf = self.psf

        if isinstance(psf, PSFImage):
            return psf._data, psf.upsample, psf.pad
        if isinstance(psf, Model):
            return psf, psf.upsample, psf.pad
        return None, 1, 0

    # Initialization functions
    ######################################################################
    @ignore_numpy_warnings
    def initialize(self):
        """Determine initial values for the center coordinates. This is done
        with a local center of mass search which iterates by finding
        the center of light in a window, then iteratively updates
        until the iterations move by less than a pixel.
        """
        if self.psf is not None and isinstance(self.psf, Model):
            self.psf.initialize()

        # Use center of window if a center hasn't been set yet
        if self.center.initialized:
            return

        target_area = self.target[self.window]
        dat = np.copy(backend.to_numpy(target_area._data))
        mask = backend.to_numpy(target_area._mask)
        dat[mask] = np.nanmedian(dat[~mask])

        COM = recursive_center_of_mass(dat)
        if not np.all(np.isfinite(COM)):
            return
        COM_center = target_area.pixel_to_plane(
            *backend.as_array(COM, dtype=config.DTYPE, device=config.DEVICE), ()
        )
        self.center.value = COM_center

    @forward
    def transform_coordinates(self, x, y, center):
        return x - center[0], y - center[1]

    @forward
    def pixel_brightness(self, i, j, _CD=None, _crtan=None, _crpix=None):
        """Evaluate the model at the pixel coordinates defined by i and j (of the target image)."""
        if _CD is None:
            x, y = self.target.pixel_to_plane(i, j)
        else:
            x, y = self.target.pixel_to_plane(i, j, CD=_CD, crtan=_crtan, _crpix=_crpix)
        return self.brightness(x, y)

    @forward
    def sample(
        self,
        I_: ArrayLike,
        J_: ArrayLike,
        psf: ArrayLike = None,
        crop: int = 0,
        downsample: int = 1,
        _CD: Optional[ArrayLike] = None,
        _crtan: Optional[ArrayLike] = None,
        _crpix: Optional[ArrayLike] = None,
    ):
        Z = self.pixel_brightness(I_, J_, _CD=_CD, _crtan=_crtan, _crpix=_crpix)
        Z = self._pixel_integrator(Z)
        I_, J_ = self._pixel_center_finder(I_, J_)
        Z = self._adaptive_integrator(
            Z,
            I_,
            J_,
            downsample,
            lambda i, j: self.pixel_brightness(i, j, _CD=_CD, _crtan=_crtan, _crpix=_crpix),
        )
        if _CD is None:
            Z = Z * self.target.pixel_collecting_area(I_, J_, downsample)
        else:
            Z = Z * self.target.pixel_collecting_area(I_, J_, downsample, CD=_CD)
        if psf is not None:
            if isinstance(psf, Model):
                psf = psf()._data
            if psf.shape != (1, 1):  # skip if identity PSF
                Z = func.convolve(Z, psf)
                Z = Z[crop : Z.shape[0] - crop, crop : Z.shape[1] - crop]
                Z = func.downsample(Z, downsample)
        return Z

    @forward
    def __call__(
        self,
        _CD: Optional[ArrayLike] = None,
        _crtan: Optional[ArrayLike] = None,
        _crpix: Optional[ArrayLike] = None,
        _psf: Optional[ArrayLike] = None,
    ) -> ModelImage:

        psf, upsample, pad = self._prep_psf()
        if _psf is not None and not isinstance(psf, Model) and self.psf_convolve:
            psf = _psf
        working_image = self.target.model_image(self.window)
        I, J = self._pixel_meshgridder(self.target, self.window, pad, upsample)

        # pixel_collecting_area: Units from flux/arcsec^2 to flux, multiply by pixel area
        working_image._data = self.sample(
            I,
            J,
            psf=psf,
            crop=pad,
            downsample=upsample,
            _CD=_CD,
            _crtan=_crtan,
            _crpix=_crpix,
        )
        return working_image



from .brightness import RadialMixin, WedgeMixin, RayMixin
from .transform import (
    InclinedMixin,
    SuperEllipseMixin,
    FourierEllipseMixin,
    WarpMixin,
    TruncationMixin,
)
from .sersic import SersicMixin, iSersicMixin, SersicPSFMixin
from .exponential import ExponentialMixin, iExponentialMixin, ExponentialPSFMixin
from .moffat import MoffatMixin, iMoffatMixin, MoffatPSFMixin
from .ferrer import FerrerMixin, iFerrerMixin, FerrerPSFMixin
from .king import KingMixin, iKingMixin, KingPSFMixin
from .gaussian import GaussianMixin, iGaussianMixin, GaussianPSFMixin
from .nuker import NukerMixin, iNukerMixin, NukerPSFMixin
from .spline import SplineMixin, iSplineMixin, SplinePSFMixin
from .sample import SampleMixin, GradMixin

__all__ = (
    "RadialMixin",
    "WedgeMixin",
    "RayMixin",
    "SuperEllipseMixin",
    "FourierEllipseMixin",
    "WarpMixin",
    "TruncationMixin",
    "InclinedMixin",
    "SersicMixin",
    "iSersicMixin",
    "SersicPSFMixin",
    "ExponentialMixin",
    "iExponentialMixin",
    "ExponentialPSFMixin",
    "MoffatMixin",
    "iMoffatMixin",
    "MoffatPSFMixin",
    "FerrerMixin",
    "iFerrerMixin",
    "FerrerPSFMixin",
    "KingMixin",
    "iKingMixin",
    "KingPSFMixin",
    "GaussianMixin",
    "iGaussianMixin",
    "GaussianPSFMixin",
    "NukerMixin",
    "iNukerMixin",
    "NukerPSFMixin",
    "SplineMixin",
    "iSplineMixin",
    "SplinePSFMixin",
    "SampleMixin",
    "GradMixin",
)



import numpy as np

from .target_image import TargetImage
from .window import Window
from .mixins import CMOSMixin
from .model_image import ModelImage
from ..backend_obj import backend


class CMOSModelImage(CMOSMixin, ModelImage):
    """A ModelImage with CMOS-specific functionality."""


class CMOSTargetImage(CMOSMixin, TargetImage):
    """
    A TargetImage with CMOS-specific functionality.
    This class is used to represent a target image with CMOS-specific features.
    It inherits from TargetImage and CMOSMixin.
    """

    def model_image(self, window: Window, **kwargs) -> CMOSModelImage:
        """Model the image with CMOS-specific features."""
        si, sj = self.get_indices(window)

        kwargs = {
            "subpixel_loc": self.subpixel_loc,
            "subpixel_scale": self.subpixel_scale,
            "_data": backend.zeros_like(self._data[si, sj]),
            "CD": self.CD.value,
            "crpix": self.crpix - np.array((si.start, sj.start)),
            "crtan": self.crtan.value,
            "crval": self.crval.value,
            "zeropoint": self.zeropoint,
            "identity": self.identity,
            "name": self.name + "_model",
            **kwargs,
        }
        return CMOSModelImage(**kwargs)



import numpy as np


def gaussian_psf(sigma, img_width, pixelscale, upsample=4, normalize=True):
    """
    create a gaussian point spread function (PSF) image.

    **Args:**
    -  `sigma`: Standard deviation of the Gaussian in arcseconds.
    -  `img_width`: Width of the PSF image in pixels.
    -  `pixelscale`: Pixel scale in arcseconds per pixel.
    -  `upsample`: Upsampling factor to more accurately create the PSF (the outputted PSF is not upsampled).
    -  `normalize`: Whether to normalize the PSF so that the sum of all pixels equals 1. If False, the PSF will not be normalized.
    """
    assert img_width % 2 == 1, "psf images should have an odd shape"

    # Number of super sampled pixels
    N = img_width * upsample
    # Grid of centered pixel locations
    XX, YY = np.meshgrid(
        np.linspace(
            -(N - 1) * pixelscale / (2 * upsample),
            (N - 1) * pixelscale / (2 * upsample),
            N,
        ),
        np.linspace(
            -(N - 1) * pixelscale / (2 * upsample),
            (N - 1) * pixelscale / (2 * upsample),
            N,
        ),
    )
    # Evaluate the Gaussian at each pixel
    ZZ = np.exp(-0.5 * (XX**2 + YY**2) / sigma**2)

    # Reduce the super-sampling back to native resolution
    ZZ = ZZ.reshape(img_width, upsample, img_width, upsample).sum(axis=(1, 3)) / (upsample**2)

    # Normalize the PSF
    if normalize:
        return ZZ / np.sum(ZZ)
    return ZZ


def moffat_psf(n, Rd, img_width, pixelscale, upsample=4, normalize=True):
    """
    Create a Moffat point spread function (PSF) image.

    **Args:**
    -  `n`: Moffat index (power-law index).
    -  `Rd`: Scale radius of the Moffat profile in arcseconds.
    -  `img_width`: Width of the PSF image in pixels.
    -  `pixelscale`: Pixel scale in arcseconds per pixel.
    -  `upsample`: Upsampling factor to more accurately create the PSF (the outputted PSF is not upsampled).
    -  `normalize`: Whether to normalize the PSF so that the sum of all pixels equals 1. If False, the PSF will not be normalized.
    """
    assert img_width % 2 == 1, "psf images should have an odd shape"

    # Number of super sampled pixels
    N = img_width * upsample
    # Grid of centered pixel locations
    XX, YY = np.meshgrid(
        np.linspace(
            -(N - 1) * pixelscale / (2 * upsample),
            (N - 1) * pixelscale / (2 * upsample),
            N,
        ),
        np.linspace(
            -(N - 1) * pixelscale / (2 * upsample),
            (N - 1) * pixelscale / (2 * upsample),
            N,
        ),
    )
    # Evaluate the Moffat at each pixel
    ZZ = 1.0 / (1.0 + (XX**2 + YY**2) / (Rd**2)) ** n

    # Reduce the super-sampling back to native resolution
    ZZ = ZZ.reshape(img_width, upsample, img_width, upsample).sum(axis=(1, 3)) / (upsample**2)

    # Normalize the PSF
    if normalize:
        return ZZ / np.sum(ZZ)
    return ZZ



from ...utils.integration import quad_table
from ...backend_obj import backend, ArrayLike


def pixel_center_meshgrid(
    extent: tuple[int, int, int, int], pad: int, upsample: int, dtype: type, device: str
) -> tuple:
    i = backend.linspace(
        extent[0] - 0.5 + 0.5 / upsample - pad / upsample,
        extent[1] - 0.5 - 0.5 / upsample + pad / upsample,
        (extent[1] - extent[0]) * upsample + 2 * pad,
        dtype=dtype,
        device=device,
    )
    j = backend.linspace(
        extent[2] - 0.5 + 0.5 / upsample - pad / upsample,
        extent[3] - 0.5 - 0.5 / upsample + pad / upsample,
        (extent[3] - extent[2]) * upsample + 2 * pad,
        dtype=dtype,
        device=device,
    )
    return backend.meshgrid(i, j, indexing="ij")


def cmos_pixel_center_meshgrid(
    extent: tuple[int, int, int, int],
    pad: int,
    upsample: int,
    loc: tuple[float, float],
    dtype,
    device,
) -> tuple:
    if upsample > 1:
        raise NotImplementedError("CMOS pixel meshgrid does not currently support upsampling.")
    i, j = pixel_center_meshgrid(extent, pad, upsample, dtype, device)
    return i + loc[0], j + loc[1]


def pixel_corner_meshgrid(
    extent: tuple[int, int, int, int], pad: int, upsample: int, dtype, device
) -> tuple:
    i, j = pixel_center_meshgrid(
        (extent[0], extent[1] + 1, extent[2], extent[3] + 1), pad, upsample, dtype, device
    )
    return i - 0.5 / upsample, j - 0.5 / upsample


def pixel_simpsons_meshgrid(
    extent: tuple[int, int, int, int], pad: int, upsample: int, dtype, device
) -> tuple:
    return pixel_corner_meshgrid(extent, 2 * pad, 2 * upsample, dtype, device)


def pixel_quad_meshgrid(
    extent: tuple[int, int, int, int], pad: int, upsample: int, dtype, device, order=3
) -> tuple:
    i, j = pixel_center_meshgrid(extent, pad, upsample, dtype, device)
    di, dj, w = quad_table(order, dtype, device)
    i = backend.repeat(i[..., None], order**2, -1) + di.flatten() / upsample
    j = backend.repeat(j[..., None], order**2, -1) + dj.flatten() / upsample
    return i, j, w.flatten()


def rotate(theta: ArrayLike, x: ArrayLike, y: ArrayLike) -> tuple:
    """
    Applies a rotation matrix to the X,Y coordinates
    """
    s = backend.sin(theta)
    c = backend.cos(theta)
    return c * x - s * y, s * x + c * y



import numpy as np
import torch

from .sky_model_object import SkyModel
from ..utils.decorators import ignore_numpy_warnings, combine_docstrings
from ..param import forward
from ..backend_obj import backend, ArrayLike

__all__ = ["PlaneSky"]


@combine_docstrings
class PlaneSky(SkyModel):
    """Sky background model using a tilted plane for the sky flux. The brightness for each pixel is defined as:

    $$I(X, Y) = I_0 + X*\\delta_x + Y*\\delta_y$$

    where $I(X,Y)$ is the brightness as a function of image position $X, Y$,
    $I_0$ is the central sky brightness value, and $\\delta_x, \\delta_y$ are the slopes of
    the sky brightness plane.

    **Parameters:**
    -    `I0`: central sky brightness value
    -    `delta`: Tensor for slope of the sky brightness in each image dimension

    """

    _model_type = "plane"
    _parameter_specs = {
        "I0": {"units": "flux/arcsec^2", "shape": (), "dynamic": True},
        "delta": {"units": "flux/arcsec", "shape": (2,), "dynamic": True},
    }
    usable = True

    @torch.no_grad()
    @ignore_numpy_warnings
    def initialize(self):
        super().initialize()

        if not self.I0.initialized:
            dat = backend.to_numpy(self.target[self.window]._data).copy()
            mask = backend.to_numpy(self.target[self.window]._mask)
            dat[mask] = np.median(dat[~mask])
            self.I0.value = np.median(dat) / self.target.pixel_area.item()
        if not self.delta.initialized:
            self.delta.value = [0.0, 0.0]

    @forward
    def brightness(self, x: ArrayLike, y: ArrayLike, I0: ArrayLike, delta: ArrayLike) -> ArrayLike:
        return I0 + x * delta[0] + y * delta[1]



from .mixins import (
    SersicPSFMixin,
    ExponentialPSFMixin,
    GaussianPSFMixin,
    FerrerPSFMixin,
    KingPSFMixin,
    MoffatPSFMixin,
    NukerPSFMixin,
    SplinePSFMixin,
    RadialMixin,
    SuperEllipseMixin,
    FourierEllipseMixin,
    WarpMixin,
    InclinedMixin,
)
from .psf_model_object import PSFModel

radial_models = (
    SersicPSFMixin,
    ExponentialPSFMixin,
    GaussianPSFMixin,
    FerrerPSFMixin,
    KingPSFMixin,
    MoffatPSFMixin,
    NukerPSFMixin,
    SplinePSFMixin,
)

EllipseMixin = type("EllipseMixin", (InclinedMixin,), {"usable": False, "_model_type": "ellipse"})
__all__ = []
for mixin in radial_models:
    # PSF Model
    g_mixin = type(mixin.__name__[:-5], (mixin, RadialMixin, PSFModel), {"usable": True})
    globals()[g_mixin.__name__] = g_mixin
    __all__.append(g_mixin.__name__)

    # Ellipse PSF Model
    g_mixin = type(
        mixin.__name__[:-5] + "Ellipse",
        (mixin, EllipseMixin, RadialMixin, PSFModel),
        {"usable": True},
    )
    globals()[g_mixin.__name__] = g_mixin
    __all__.append(g_mixin.__name__)

    for n, p in zip(
        ("SuperEllipse", "FourierEllipse", "Warp"),
        (SuperEllipseMixin, FourierEllipseMixin, WarpMixin),
    ):
        # Galaxy Model with additional perturbation mixin
        g_mixin = type(
            mixin.__name__[:-5] + n,
            (mixin, InclinedMixin, RadialMixin, p, PSFModel),
            {"usable": True},
        )
        globals()[g_mixin.__name__] = g_mixin
        __all__.append(g_mixin.__name__)



# Traditional gradient descent with Adam
from time import time
from typing import Sequence
from caskade import ValidContext
import torch
import numpy as np

from .base import BaseOptimizer
from .. import config
from ..backend_obj import backend, ArrayLike
from ..models import Model
from ..errors import OptimizeStopFail, OptimizeStopSuccess
from . import func
from ..utils.decorators import combine_docstrings

__all__ = ["Grad"]


@combine_docstrings
class Grad(BaseOptimizer):
    """A gradient descent optimization wrapper for AstroPhot Model objects.

    The default method is "NAdam", a variant of the Adam optimization algorithm.
    This optimizer uses a combination of gradient descent and Nesterov momentum for faster convergence.
    The optimizer is instantiated with a set of initial parameters and optimization options provided by the user.
    The `fit` method performs the optimization, taking a series of gradient steps until a stopping criteria is met.

    **Args:**
    -  `likelihood` (str, optional): The likelihood function to use for the optimization. Defaults to "gaussian".
    -  `method` (str, optional): the optimization method to use for the update step. Defaults to "NAdam".
    -  `optim_kwargs` (dict, optional): a dictionary of keyword arguments to pass to the pytorch optimizer.
    -  `patience` (int, optional): number of steps with no improvement before stopping the optimization. Defaults to 10.
    -  `report_freq` (int, optional): frequency of reporting the optimization progress. Defaults to 10 steps.
    """

    def __init__(
        self,
        model: Model,
        initial_state: Sequence = None,
        likelihood="gaussian",
        method="NAdam",
        optim_kwargs={},
        patience: int = 10,
        report_freq=10,
        **kwargs,
    ) -> None:

        super().__init__(model, initial_state, **kwargs)

        self.likelihood = likelihood

        # set parameters from the user
        self.patience = patience
        self.method = method
        self.optim_kwargs = optim_kwargs
        self.report_freq = report_freq

        # Default learning rate if none given. Equal to 1 / sqrt(parames)
        if "lr" not in self.optim_kwargs:
            self.optim_kwargs["lr"] = 0.1 / (len(self.current_state) ** (0.5))

        # Instantiates the appropriate pytorch optimizer with the initial state and user provided kwargs
        self.current_state.requires_grad = True
        self.optimizer = getattr(torch.optim, self.method)(
            (self.current_state,), **self.optim_kwargs
        )

    def density(self, state: torch.Tensor) -> torch.Tensor:
        """
        Returns the density of the model at the given state vector. This is used
        to calculate the likelihood of the model at the given state. Based on
        ``self.likelihood``, will be either the Gaussian or Poisson negative log
        likelihood.
        """
        if self.likelihood == "gaussian":
            return -self.model.gaussian_log_likelihood(state)
        elif self.likelihood == "poisson":
            return -self.model.poisson_log_likelihood(state)
        else:
            raise ValueError(f"Unknown likelihood type: {self.likelihood}")

    def step(self) -> None:
        """Take a single gradient step.

        Computes the loss function of the model,
        computes the gradient of the parameters using automatic differentiation,
        and takes a step with the PyTorch optimizer.

        """
        self.iteration += 1

        self.optimizer.zero_grad()
        self.current_state.requires_grad = True
        loss = self.density(self.current_state)

        loss.backward()

        self.loss_history.append(backend.to_numpy(loss))
        self.lambda_history.append(np.copy(backend.to_numpy(self.current_state)))
        if (
            self.iteration % int(self.max_iter / self.report_freq) == 0
        ) or self.iteration == self.max_iter:
            if self.verbose > 0:
                config.logger.info(f"iter: {self.iteration}, posterior density: {loss.item():.6e}")
            if self.verbose > 1:
                config.logger.info(f"gradient: {self.current_state.grad}")
        self.optimizer.step()

    def fit(self) -> BaseOptimizer:
        """
        Perform an iterative fit of the model parameters using the specified optimizer.

        The fit procedure continues until a stopping criteria is met,
        such as the maximum number of iterations being reached,
        or no improvement being made after a specified number of iterations.

        """
        start_fit = time()
        try:
            while True:
                self.step()
                if self.iteration >= self.max_iter:
                    self.message = self.message + " fail max iteration reached"
                    break
                if (
                    self.patience is not None
                    and (len(self.loss_history) - np.argmin(self.loss_history)) > self.patience
                ):
                    self.message = self.message + " fail no improvement"
                    break
                L = np.sort(self.loss_history)
                if len(L) >= 5 and 0 < (L[4] - L[0]) / L[0] < self.relative_tolerance:
                    self.message = self.message + " success"
                    break
        except KeyboardInterrupt:
            self.message = self.message + " fail interrupted"

        # Set the model parameters to the best values from the fit and clear any previous model sampling
        self.model.set_values(torch.tensor(self.res(), dtype=config.DTYPE, device=config.DEVICE))
        if self.verbose > 1:
            config.logger.info(
                f"Grad Fitting complete in {time() - start_fit} sec with message: {self.message}"
            )
        return self


class Slalom(BaseOptimizer):
    """Slalom optimizer for Model objects.

    Slalom is a gradient descent optimization algorithm that uses a few
    evaluations along the direction of the gradient to find the optimal step
    size. This is done by assuming that the posterior density is a parabola and
    then finding the minimum.

    The optimizer quickly finds the minimum of the posterior density along the
    gradient direction, then updates the gradient at the new position and
    repeats. This continues until it reaches a set of 5 steps which collectively
    improve the posterior density by an amount smaller than the
    `relative_tolerance` threshold, indicating that convergence has been
    achieved. Note that this convergence criteria is not a guarantee, simply a
    heuristic. The default tolerance was such that the optimizer will
    substantially improve from the starting point, and do so quickly, but may
    not reach all the way to the minimum of the posterior density. Like other
    gradient descent algorithms, Slalom slows down considerably when trying to
    achieve very high precision.

    **Args:**
    -  `S` (float, optional): The initial step size for the Slalom optimizer. Defaults to 1e-4.
    -  `likelihood` (str, optional): The likelihood function to use for the optimization. Defaults to "gaussian".
    -  `report_freq` (int, optional): Frequency of reporting the optimization progress. Defaults to 10 steps.
    -  `relative_tolerance` (float, optional): The relative tolerance for convergence. Defaults to 1e-4.
    -  `momentum` (float, optional): The momentum factor for the Slalom optimizer. Defaults to 0.5.
    -  `max_iter` (int, optional): The maximum number of iterations for the optimizer. Defaults to 1000.
    """

    def __init__(
        self,
        model: Model,
        initial_state: Sequence = None,
        S=1e-4,
        likelihood: str = "gaussian",
        report_freq: int = 10,
        relative_tolerance: float = 1e-4,
        momentum: float = 0.5,
        max_iter: int = 1000,
        **kwargs,
    ) -> None:
        """Initialize the Slalom optimizer."""
        super().__init__(
            model, initial_state, relative_tolerance=relative_tolerance, max_iter=max_iter, **kwargs
        )
        self.likelihood = likelihood
        self.S = S
        self.report_freq = report_freq
        self.momentum = momentum

    def density(self, state: ArrayLike) -> ArrayLike:
        """Calculate the density of the model at the given state. Based on
        ``self.likelihood``, will be either the Gaussian or Poisson negative log
        likelihood."""
        if self.likelihood == "gaussian":
            return -self.model.gaussian_log_likelihood(state)
        elif self.likelihood == "poisson":
            return -self.model.poisson_log_likelihood(state)
        else:
            raise ValueError(f"Unknown likelihood type: {self.likelihood}")

    def fit(self) -> BaseOptimizer:
        """Perform the Slalom optimization."""

        grad_func = backend.grad(self.density)
        momentum = backend.zeros_like(self.current_state)
        self.S_history = [self.S]
        self.loss_history = [self.density(self.current_state).item()]
        self.lambda_history = [backend.to_numpy(self.current_state)]
        self.start_fit = time()

        for i in range(self.max_iter):

            try:
                # Perform the Slalom step
                vstate = self.model.to_valid(self.current_state)
                with ValidContext(self.model):
                    self.S, loss, grad = func.slalom_step(
                        self.density, grad_func, vstate, m=momentum, S=self.S
                    )
                self.current_state = self.model.from_valid(
                    vstate - self.S * (grad + momentum) / backend.linalg.norm(grad + momentum)
                )
                momentum = self.momentum * (momentum + grad)
            except OptimizeStopSuccess as e:
                self.message = self.message + str(e)
                break
            except OptimizeStopFail as e:
                if backend.allclose(momentum, backend.zeros_like(momentum)):
                    self.message = self.message + str(e)
                    break
                momentum = backend.zeros_like(self.current_state)
                continue
            # Log the loss
            self.S_history.append(self.S)
            self.loss_history.append(loss)
            self.lambda_history.append(backend.to_numpy(self.current_state))

            if self.verbose > 0 and (i % int(self.report_freq) == 0 or i == self.max_iter - 1):
                config.logger.info(
                    f"iter: {i}, step size: {self.S:.6e}, posterior density: {loss:.6e}"
                )

            if len(self.loss_history) >= 5:
                relative_loss = (self.loss_history[-5] - self.loss_history[-1]) / self.loss_history[
                    -1
                ]
                if relative_loss < self.relative_tolerance:
                    self.message = self.message + " success"
                    break
        else:
            self.message = self.message + " fail. max iteration reached"

        # Set the model parameters to the best values from the fit
        self.model.set_values(
            backend.as_array(self.res(), dtype=config.DTYPE, device=config.DEVICE)
        )
        if self.verbose > 0:
            config.logger.info(
                f"Slalom Fitting complete in {time() - self.start_fit} sec with message: {self.message}"
            )
        return self



import os
import astrophot as ap
import numpy as np
from utils import make_basic_sersic
import pytest

# torch.autograd.set_detect_anomaly(True)
######################################################################
# Model Objects
######################################################################


def test_model_sampling_modes():

    target = make_basic_sersic(90, 100)
    model = ap.Model(
        name="test_sersic",
        model_type="sersic galaxy model",
        center=[40, 41.9],
        PA=60 * np.pi / 180,
        q=0.8,
        n=0.5,
        Re=20,
        Ie=1,
        target=target,
    )

    # With subpixel integration
    model.integrate_mode = "bright"
    auto = ap.backend.to_numpy(model().data)
    model.sampling_mode = "midpoint"
    midpoint = ap.backend.to_numpy(model().data)
    midpoint_bright = midpoint.copy()
    model.sampling_mode = "simpsons"
    simpsons = ap.backend.to_numpy(model().data)
    model.sampling_mode = "upsample:5"
    upsample5 = ap.backend.to_numpy(model().data)
    model.sampling_mode = "quad:5"
    quad5 = ap.backend.to_numpy(model().data)
    assert np.allclose(midpoint, auto, rtol=1e-2), "Midpoint sampling should match auto sampling"
    assert np.allclose(midpoint, simpsons, rtol=1e-2), "Simpsons sampling should match midpoint"
    assert np.allclose(midpoint, quad5, rtol=1e-2), "Quad5 sampling should match midpoint sampling"
    assert np.allclose(
        midpoint, upsample5, rtol=1e-2
    ), "Upsample5 sampling should match midpoint sampling"
    assert np.allclose(simpsons, quad5, rtol=1e-6), "Quad5 sampling should match Simpsons sampling"

    # Without subpixel integration
    model.integrate_mode = "none"
    auto = ap.backend.to_numpy(model().data)
    model.sampling_mode = "midpoint"
    midpoint = ap.backend.to_numpy(model().data)
    model.sampling_mode = "simpsons"
    simpsons = ap.backend.to_numpy(model().data)
    model.sampling_mode = "upsample:5"
    upsample5 = ap.backend.to_numpy(model().data)
    model.sampling_mode = "quad:5"
    quad5 = ap.backend.to_numpy(model().data)
    assert np.allclose(
        midpoint, midpoint_bright, rtol=1e-2
    ), "no integrate sampling should match bright sampling"
    assert np.allclose(midpoint, auto, rtol=1e-2), "Midpoint sampling should match auto sampling"
    assert np.allclose(midpoint, simpsons, rtol=1e-2), "Simpsons sampling should match midpoint"
    assert np.allclose(midpoint, quad5, rtol=1e-2), "Quad5 sampling should match midpoint sampling"
    assert np.allclose(
        midpoint, upsample5, rtol=1e-2
    ), "Upsample5 sampling should match midpoint sampling"
    assert np.allclose(simpsons, quad5, rtol=1e-6), "Quad5 sampling should match Simpsons sampling"

    # curvature based subpixel integration
    model.integrate_mode = "curvature"
    auto = ap.backend.to_numpy(model().data)
    model.sampling_mode = "midpoint"
    midpoint = ap.backend.to_numpy(model().data)
    model.sampling_mode = "simpsons"
    simpsons = ap.backend.to_numpy(model().data)
    model.sampling_mode = "upsample:5"
    upsample5 = ap.backend.to_numpy(model().data)
    model.sampling_mode = "quad:5"
    quad5 = ap.backend.to_numpy(model().data)
    assert np.allclose(
        midpoint, midpoint_bright, rtol=1e-2
    ), "curvature integrate sampling should match bright sampling"
    assert np.allclose(midpoint, auto, rtol=1e-2), "Midpoint sampling should match auto sampling"
    assert np.allclose(midpoint, simpsons, rtol=1e-2), "Simpsons sampling should match midpoint"
    assert np.allclose(midpoint, quad5, rtol=1e-2), "Quad5 sampling should match midpoint sampling"
    assert np.allclose(
        midpoint, upsample5, rtol=1e-2
    ), "Upsample5 sampling should match midpoint sampling"
    assert np.allclose(simpsons, quad5, rtol=1e-6), "Quad5 sampling should match Simpsons sampling"

    with pytest.raises(ap.errors.SpecificationConflict):
        model.integrate_mode = "should raise"
    model.integrate_mode = "none"
    with pytest.raises(ap.errors.SpecificationConflict):
        model.sampling_mode = "should raise"
    model.sampling_mode = "midpoint"
    model.integrate_mode = "none"

    # test PSF modes
    model.psf = np.array([[0.05, 0.1, 0.05], [0.1, 0.4, 0.1], [0.05, 0.1, 0.05]])
    model.psf_convolve = True
    model()


def test_model_errors():

    # Target that is not a target image
    arr = np.zeros((10, 15))
    target = ap.image.Image(data=arr, pixelscale=1.0, zeropoint=1.0)

    with pytest.raises(ap.errors.InvalidTarget):
        ap.Model(
            name="test_model",
            model_type="sersic galaxy model",
            target=target,
        )

    # model that doesn't exist
    target = make_basic_sersic()
    with pytest.raises(ap.errors.UnrecognizedModel):
        ap.Model(
            name="test_model",
            model_type="sersic gaaxy model",
            target=target,
        )


@pytest.mark.parametrize(
    "model_type", ap.models.ComponentModel.List_Models(usable=True, types=True)
)
def test_all_model_sample(model_type):

    if ap.backend.backend == "jax" and np.random.randint(0, 3) > 0:
        pytest.skip("JAX is very slow, randomly reducing the number of tests")
    if model_type == "isothermal sech2 edgeon model" and ap.backend.backend == "jax":
        pytest.skip("JAX doesnt have bessel function k1 yet")

    if (
        model_type in ["ferrer warp galaxy model", "king warp galaxy model"]
        and ap.backend.backend == "jax"
    ):
        pytest.skip("JAX version doesnt support these models yet, difficulty with gradients")

    if any(t in model_type for t in ["warp", "fourier"]):
        pytest.skip("Warp and Fourier models are complex and slow to fit, skipping for now")
    if model_type.startswith("truncated"):
        pytest.skip("Testing truncated models is redundant")

    target = make_basic_sersic()
    target.zeropoint = 22.5
    MODEL = ap.Model(
        name="test_model",
        model_type=model_type,
        target=target,
        integrate_mode=(
            "none" if ap.backend.backend == "jax" else "bright"
        ),  # JAX JIT is reallly slow for any integration
    )
    MODEL.initialize()
    MODEL.to()
    for P in MODEL.dynamic_params:
        assert (
            P.value is not None
        ), f"Model type {model_type} parameter {P.name} should not be None after initialization"
    img = MODEL()
    assert ap.backend.all(
        ap.backend.isfinite(img.data)
    ), "Model should evaluate a real number for the full image"

    res = ap.fit.LM(MODEL, max_iter=5, verbose=1).fit()

    print(MODEL)  # test printing

    # sky has little freedom to fit, some more complex models need extra
    # attention to get a good fit so here we just check that they can improve
    if (
        "sky" in model_type
        or "king" in model_type
        or "spline" in model_type
        or model_type
        in [
            "exponential warp galaxy model",
            "ferrer warp galaxy model",
            "ferrer ray galaxy model",
            "isothermal sech2 edgeon model",
        ]
    ):
        assert res.loss_history[0] > res.loss_history[-1], (
            f"Model {model_type} should fit to the target image, but did not. "
            f"Initial loss: {res.loss_history[0]}, Final loss: {res.loss_history[-1]}"
        )
    else:  # Most models should get significantly better after just a few iterations
        assert res.loss_history[0] > (1.5 * res.loss_history[-1]), (
            f"Model {model_type} should fit to the target image, but did not. "
            f"Initial loss: {res.loss_history[0]}, Final loss: {res.loss_history[-1]}"
        )

    F = MODEL.total_flux()
    assert ap.backend.isfinite(F), "Model total flux should be finite after fitting"
    assert F > 0, "Model total flux should be positive after fitting"
    U = MODEL.total_flux_uncertainty()
    assert ap.backend.isfinite(U), "Model total flux uncertainty should be finite after fitting"
    assert U >= 0, "Model total flux uncertainty should be non-negative after fitting"
    M = MODEL.total_magnitude()
    assert ap.backend.isfinite(M), "Model total magnitude should be finite after fitting"
    U_M = MODEL.total_magnitude_uncertainty()
    assert ap.backend.isfinite(
        U_M
    ), "Model total magnitude uncertainty should be finite after fitting"
    assert U_M >= 0, "Model total magnitude uncertainty should be non-negative after fitting"

    allnames = set()
    for name in MODEL.build_params_array_names():
        assert name not in allnames, f"Duplicate parameter name found: {name}"
        allnames.add(name)


def test_sersic_save_load():

    target = make_basic_sersic()
    model = ap.Model(
        name="test_sersic",
        model_type="sersic galaxy model",
        center=[20, 20],
        PA=60 * np.pi / 180,
        q=0.5,
        n=2,
        Re=5,
        Ie=1,
        target=target,
    )

    model.initialize()
    model.save_state("test_AstroPhot_sersic.hdf5", appendable=True)
    model.center = [30, 30]
    model.PA = 30 * np.pi / 180
    model.q = 0.8
    model.n = 3
    model.Re = 10
    model.Ie = 2
    target.crtan = [1.0, 2.0]
    model.append_state("test_AstroPhot_sersic.hdf5")
    model.load_state("test_AstroPhot_sersic.hdf5", index=0)

    assert model.center.value[0].item() == 20, "Model center should be loaded correctly"
    assert model.center.value[1].item() == 20, "Model center should be loaded correctly"
    assert model.PA.value.item() == 60 * np.pi / 180, "Model PA should be loaded correctly"
    assert model.q.value.item() == 0.5, "Model q should be loaded correctly"
    assert model.n.value.item() == 2, "Model n should be loaded correctly"
    assert model.Re.value.item() == 5, "Model Re should be loaded correctly"
    assert model.Ie.value.item() == 1, "Model Ie should be loaded correctly"
    assert model.target.crtan.value[0] == 0.0, "Model target crtan should be loaded correctly"
    assert model.target.crtan.value[1] == 0.0, "Model target crtan should be loaded correctly"

    os.remove("test_AstroPhot_sersic.hdf5")


def test_batch_model(sersic):
    M = ap.Model(model_type="batch model", model=sersic)
    M.initialize()
    # Now any parameters may be given an extra dimension for the batch
    sersic.center = [[10, 10], [20, 70], [50, 50], [70, 30], [80, 80]]
    sersic.q = [0.7, 0.6, 0.5, 0.4, 0.3]
    sersic.PA = np.array([30, 60, 90, 120, 150]) * np.pi / 180

    img = M()
    assert isinstance(img, ap.image.ModelImage), "Output of batch model should be a ModelImage"
    assert img.data.shape == sersic.target.data.shape, "batched model should produce regular output"

    M.set_values(ap.backend.ones_like(M.get_values()), attribute="uncertainty")

    assert M.total_flux() > 0, "Total flux of batch model should be positive"
    assert (
        M.total_flux_uncertainty() > 0
    ), "Total flux uncertainty of batch model should be positive"


def test_batch_scene_model(sersic):
    target_batch = ap.TargetImageBatch(
        images=[
            ap.TargetImage(
                data=np.zeros((52, 50)),
                CD=ap.utils.initialize.R(np.pi / 3),
                crtan=(10, 0),
                name="image1",
            ),
            ap.TargetImage(
                data=np.zeros((52, 50)), CD=ap.utils.initialize.R(np.pi / 6), name="image2"
            ),
            ap.TargetImage(
                data=np.zeros((52, 50)),
                CD=ap.utils.initialize.R(np.pi / 16),
                crtan=(-15, 0),
                name="image3",
            ),
        ]
    )
    model = ap.Model(model_type="batch scene model", model=sersic, target=target_batch)
    model.initialize()
    sersic.Ie = [10, 20, 30]
    img = model()
    assert isinstance(
        img, ap.image.ModelImageBatch
    ), "Output of batch scene model should be a ModelImageBatch"
    assert img.data.shape[0] == 3, "Batch scene model should produce output for each target image"
    jac = model.jacobian()
    assert isinstance(
        jac, ap.image.JacobianImageBatch
    ), "Jacobian of batch scene model should be a JacobianImageBatch"
    assert (
        jac.data.shape[0] == 3
    ), "Batch scene model jacobian should produce output for each target image"



import torch
import numpy as np

from .model_object import ComponentModel
from ..utils.decorators import ignore_numpy_warnings, combine_docstrings
from ..utils.interpolate import interp2d
from ..param import forward
from ..backend_obj import backend, ArrayLike
from ..utils.initialize import polar_decomposition
from . import func

__all__ = ["Pixelated"]


@combine_docstrings
class Pixelated(ComponentModel):
    """Model described by a grid of pixels on the sky (bilinear interpolation)

    This model represents an extended source on the sky as a grid of pixels with
    some brightness. The brightness at any point is determined by bilinear
    interpolation of the grid values. This is a very flexible model that can
    represent any source, but it is also computationally expensive to optimize
    the large number of free parameters.

    The PA and scale are also parameters of this model, so one could alternately
    fix the pixels to some image and just fit the PA and scale.

    **Parameters:**
        - `I`: the total flux within each pixel, represented as the log of the flux.
        - `PA`: the position angle of the model, in radians.
        - `scale`: the scale of the model, in arcsec per grid unit.

    """

    _model_type = "pixelated"
    _parameter_specs = {
        "I": {"units": "flux/arcsec^2", "shape": (None, None), "dynamic": True},
        "PA": {"units": "radians", "shape": (), "dynamic": False},
        "scale": {"units": "arcsec/grid-unit", "shape": (), "dynamic": False},
    }
    usable = True

    def __init__(
        self, *args, scale=None, sampling_mode="upsample:3", integrate_mode="none", **kwargs
    ):
        super().__init__(
            *args, sampling_mode=sampling_mode, integrate_mode=integrate_mode, **kwargs
        )
        self.scale = scale

    @torch.no_grad()
    @ignore_numpy_warnings
    def initialize(self):
        super().initialize()
        if not self.scale.initialized:
            self.scale = 2 * self.target.pixelscale.item()

        if not self.I.initialized:
            target_area = self.target[self.window]
            self.I.value = func.downsample(backend.copy(target_area.data), 2) / (
                target_area.pixel_area * 4
            )

        if not self.PA.initialized:
            R, _ = polar_decomposition(self.target.CD.npvalue)
            self.PA.value = np.arccos(np.abs(R[0, 0]))

    @forward
    def transform_coordinates(
        self, x: ArrayLike, y: ArrayLike, PA: ArrayLike, scale: ArrayLike
    ) -> tuple[ArrayLike, ArrayLike]:
        x, y = super().transform_coordinates(x, y)
        x, y = func.rotate(-PA + np.pi / 2, x, y)
        return x / scale, y / scale

    @forward
    def brightness(self, x: ArrayLike, y: ArrayLike, I: ArrayLike) -> ArrayLike:
        x, y = self.transform_coordinates(x, y)
        pixel_center = (I.shape[0] - 1) / 2, (I.shape[1] - 1) / 2
        return interp2d(I, x + pixel_center[0], y + pixel_center[1])



from .image import (
    pixel_center_meshgrid,
    cmos_pixel_center_meshgrid,
    pixel_corner_meshgrid,
    pixel_simpsons_meshgrid,
    pixel_quad_meshgrid,
    rotate,
)
from .wcs import (
    world_to_plane_gnomonic,
    plane_to_world_gnomonic,
    pixel_to_plane_linear,
    plane_to_pixel_linear,
    sip_delta,
    sip_coefs,
    sip_backward_transform,
    sip_matrix,
)

__all__ = (
    "pixel_center_meshgrid",
    "cmos_pixel_center_meshgrid",
    "pixel_corner_meshgrid",
    "pixel_simpsons_meshgrid",
    "pixel_quad_meshgrid",
    "rotate",
    "world_to_plane_gnomonic",
    "plane_to_world_gnomonic",
    "pixel_to_plane_linear",
    "plane_to_pixel_linear",
    "sip_delta",
    "sip_coefs",
    "sip_backward_transform",
    "sip_matrix",
)



from ...backend_obj import backend, ArrayLike
from .sersic import sersic_n_to_b

b = sersic_n_to_b(1.0)


def exponential(R: ArrayLike, Re: ArrayLike, Ie: ArrayLike) -> ArrayLike:
    """Exponential 1d profile function, specifically designed for pytorch
    operations.

    **Args:**
    -  `R`: Radius tensor at which to evaluate the exponential function
    -  `Re`: Effective radius in the same units as R
    -  `Ie`: Effective surface density
    """
    return Ie * backend.exp(-b * ((R / Re) - 1.0))



from .functions import (
    sersic_n_to_b,
    sersic_I0_to_flux_np,
    sersic_flux_to_I0_np,
    sersic_Ie_to_flux_np,
    sersic_flux_to_Ie_np,
    sersic_I0_to_flux_torch,
    sersic_flux_to_I0_torch,
    sersic_Ie_to_flux_torch,
    sersic_flux_to_Ie_torch,
    sersic_inv_np,
    sersic_inv_torch,
    moffat_I0_to_flux,
)
from .units import (
    deg_to_arcsec,
    arcsec_to_deg,
    flux_to_sb,
    flux_to_mag,
    sb_to_flux,
    mag_to_flux,
    magperarcsec2_to_mag,
    mag_to_magperarcsec2,
)

__all__ = (
    "sersic_n_to_b",
    "sersic_I0_to_flux_np",
    "sersic_flux_to_I0_np",
    "sersic_Ie_to_flux_np",
    "sersic_flux_to_Ie_np",
    "sersic_I0_to_flux_torch",
    "sersic_flux_to_I0_torch",
    "sersic_Ie_to_flux_torch",
    "sersic_flux_to_Ie_torch",
    "sersic_inv_np",
    "sersic_inv_torch",
    "moffat_I0_to_flux",
    "deg_to_arcsec",
    "arcsec_to_deg",
    "flux_to_sb",
    "flux_to_mag",
    "sb_to_flux",
    "mag_to_flux",
    "magperarcsec2_to_mag",
    "mag_to_magperarcsec2",
)



from .profile import (
    radial_light_profile,
    radial_median_profile,
    ray_light_profile,
    warp_phase_profile,
)
from .image import target_image, model_image, residual_image, model_window, psf_image
from .visuals import main_pallet, cmap_div, cmap_grad
from .diagnostic import covariance_matrix

__all__ = (
    "radial_light_profile",
    "radial_median_profile",
    "ray_light_profile",
    "warp_phase_profile",
    "target_image",
    "model_image",
    "residual_image",
    "model_window",
    "psf_image",
    "main_pallet",
    "cmap_div",
    "cmap_grad",
    "covariance_matrix",
)



import numpy as np
from ...backend_obj import backend
from ... import config

deg_to_rad = np.pi / 180
rad_to_deg = 180 / np.pi
rad_to_arcsec = rad_to_deg * 3600
arcsec_to_rad = deg_to_rad / 3600


def world_to_plane_gnomonic(ra, dec, ra0, dec0, x0=0.0, y0=0.0):
    """
    Convert world coordinates (RA, Dec) to plane coordinates (x, y) using the gnomonic projection.

    **Args:**
    - `ra`: (torch.Tensor) Right Ascension in degrees.
    - `dec`: (torch.Tensor) Declination in degrees.
    - `ra0`: (torch.Tensor) Reference Right Ascension in degrees.
    - `dec0`: (torch.Tensor) Reference Declination in degrees.

    **Returns:**
    - `x`: (torch.Tensor) x coordinate in arcseconds.
    - `y`: (torch.Tensor) y coordinate in arcseconds.
    """
    ra = ra * deg_to_rad
    dec = dec * deg_to_rad
    ra0 = ra0 * deg_to_rad
    dec0 = dec0 * deg_to_rad

    cosc = backend.sin(dec0) * backend.sin(dec) + backend.cos(dec0) * backend.cos(
        dec
    ) * backend.cos(ra - ra0)

    x = backend.cos(dec) * backend.sin(ra - ra0)

    y = backend.cos(dec0) * backend.sin(dec) - backend.sin(dec0) * backend.cos(dec) * backend.cos(
        ra - ra0
    )

    return x * rad_to_arcsec / cosc + x0, y * rad_to_arcsec / cosc + y0


def plane_to_world_gnomonic(x, y, ra0, dec0, x0=0.0, y0=0.0, s=1e-10):
    """
    Convert plane coordinates (x, y) to world coordinates (RA, Dec) using the gnomonic projection.

    **Args:**
    - `x`: (Tensor) x coordinate in arcseconds.
    - `y`: (Tensor) y coordinate in arcseconds.
    - `ra0`: (Tensor) Reference Right Ascension in degrees.
    - `dec0`: (Tensor) Reference Declination in degrees.
    - `s`: (float) Small constant to avoid division by zero.

    **Returns:**
    - `ra`: (Tensor) Right Ascension in degrees.
    - `dec`: (Tensor) Declination in degrees.
    """
    x = (x - x0) * arcsec_to_rad
    y = (y - y0) * arcsec_to_rad
    ra0 = ra0 * deg_to_rad
    dec0 = dec0 * deg_to_rad

    rho = backend.sqrt(x**2 + y**2) + s
    c = backend.arctan(rho)

    ra = ra0 + backend.arctan2(
        x * backend.sin(c),
        rho * backend.cos(dec0) * backend.cos(c) - y * backend.sin(dec0) * backend.sin(c),
    )

    dec = backend.arcsin(
        backend.cos(c) * backend.sin(dec0) + y * backend.sin(c) * backend.cos(dec0) / rho
    )

    return ra * rad_to_deg, dec * rad_to_deg


def pixel_to_plane_linear(i, j, i0, j0, CD, x0=0.0, y0=0.0):
    """
    Convert pixel coordinates to a tangent plane using the WCS information. This
    matches the FITS convention for linear transformations.

    **Args:**
    -  `i` (Tensor): The first coordinate of the pixel in pixel units.
    -  `j` (Tensor): The second coordinate of the pixel in pixel units.
    -  `i0` (Tensor): The i reference pixel coordinate in pixel units.
    -  `j0` (Tensor): The j reference pixel coordinate in pixel units.
    -  `CD` (Tensor): The CD matrix in arcsec per pixel. This 2x2 matrix is used to convert
       from pixel to arcsec units and also handles rotation/skew.
    -  `x0` (float): The x reference coordinate in arcseconds.
    -  `y0` (float): The y reference coordinate in arcseconds.

    **Returns:**
    -  Tuple[Tensor, Tensor]: Tuple containing the x and y coordinates in arcseconds
    """
    uv = backend.stack((i.flatten() - i0, j.flatten() - j0), dim=0)
    xy = CD @ uv

    return xy[0].reshape(i.shape) + x0, xy[1].reshape(i.shape) + y0


def sip_coefs(order):
    coefs = []
    for p in range(order + 1):
        for q in range(order + 1 - p):
            coefs.append((p, q))
    return tuple(coefs)


def sip_matrix(u, v, order):
    M = backend.zeros(
        (len(u), (order + 1) * (order + 2) // 2), dtype=config.DTYPE, device=config.DEVICE
    )
    for i, (p, q) in enumerate(sip_coefs(order)):
        M = backend.fill_at_indices(M, (slice(None), i), u**p * v**q)
    return M


def sip_backward_transform(u, v, U, V, A_ORDER, B_ORDER):
    """
    Credit: Shu Liu and Lei Hi, see here:
    https://github.com/Roman-Supernova-PIT/sfft/blob/master/sfft/utils/CupyWCSTransform.py

    Compute the backward transformation from (U, V) to (u, v)
    """

    FP_UV = sip_matrix(U, V, A_ORDER)
    GP_UV = sip_matrix(U, V, B_ORDER)

    AP = backend.linalg.lstsq(FP_UV, (u.flatten() - U).reshape(-1, 1))[0].squeeze(1)
    BP = backend.linalg.lstsq(GP_UV, (v.flatten() - V).reshape(-1, 1))[0].squeeze(1)
    return AP, BP


def sip_delta(u, v, sipA=(), sipB=()):
    """
    u = j - j0
    v = i - i0
    sipA = dict(tuple(int,int), float)
        The SIP coefficients, where the keys are tuples of powers (i, j) and the values are the coefficients.
        For example, {(1, 2): 0.1} means delta_u = 0.1 * (u * v^2).
    """
    delta_u = backend.zeros_like(u)
    delta_v = backend.zeros_like(v)
    # Get all used coefficient powers
    all_a = set(s[0] for s in sipA) | set(s[0] for s in sipB)
    all_b = set(s[1] for s in sipA) | set(s[1] for s in sipB)
    # Pre-compute all powers of u and v
    u_a = dict((a, u**a) for a in all_a)
    v_b = dict((b, v**b) for b in all_b)
    for a, b in sipA:
        delta_u = delta_u + sipA[(a, b)] * (u_a[a] * v_b[b])
    for a, b in sipB:
        delta_v = delta_v + sipB[(a, b)] * (u_a[a] * v_b[b])
    return delta_u, delta_v


def plane_to_pixel_linear(x, y, i0, j0, CD, x0=0.0, y0=0.0):
    """
    Convert tangent plane coordinates to pixel coordinates using the WCS
    information. This matches the FITS convention for linear transformations.

    **Args:**
    - `x`: (Tensor) The first coordinate of the pixel in arcsec.
    - `y`: (Tensor) The second coordinate of the pixel in arcsec.
    - `i0`: (Tensor) The i reference pixel coordinate in pixel units.
    - `j0`: (Tensor) The j reference pixel coordinate in pixel units.
    - `CD`: (Tensor) The CD matrix in arcsec per pixel.
    - `x0`: (float) The x reference coordinate in arcsec.
    - `y0`: (float) The y reference coordinate in arcsec.

    **Returns:**
    -  Tuple[Tensor, Tensor]: Tuple containing the i and j pixel coordinates in pixel units.
    """
    xy = backend.stack((x.flatten() - x0, y.flatten() - y0), dim=0)
    uv = backend.linalg.inv(CD) @ xy

    return uv[0].reshape(x.shape) + i0, uv[1].reshape(y.shape) + j0



import torch
from ...backend_obj import backend, ArrayLike


def ferrer(
    R: ArrayLike, rout: ArrayLike, alpha: ArrayLike, beta: ArrayLike, I0: ArrayLike
) -> ArrayLike:
    """
    Modified Ferrer profile.

    **Args:**
    -  `R`: Radius tensor at which to evaluate the modified Ferrer function
    -  `rout`: Outer radius of the profile
    -  `alpha`: Power-law index
    -  `beta`: Exponent for the modified Ferrer function
    -  `I0`: Central intensity
    """
    return backend.where(
        R < rout,
        I0 * ((1 - (backend.clamp(R, 0, rout) / rout) ** (2 - beta)) ** alpha),
        backend.zeros_like(R),
    )



from .model_object import ComponentModel
from .mixins import InclinedMixin
from ..utils.decorators import combine_docstrings

__all__ = ("GalaxyModel",)


@combine_docstrings
class GalaxyModel(InclinedMixin, ComponentModel):
    """Intended to represent a galaxy or extended component in an image."""

    _model_type = "galaxy"
    usable = False



import numpy as np
import torch

from ..utils.decorators import ignore_numpy_warnings, combine_docstrings
from .sky_model_object import SkyModel
from ..backend_obj import backend, ArrayLike
from ..param import forward

__all__ = ["FlatSky"]


@combine_docstrings
class FlatSky(SkyModel):
    """Model for the sky background in which all values across the image
    are the same.

    **Parameters:**
    -    `I0`: brightness for the sky, represented as the log of the brightness over pixel scale squared, this is proportional to a surface brightness

    """

    _model_type = "flat"
    _parameter_specs = {"I0": {"units": "flux/arcsec^2", "shape": (), "dynamic": True}}
    usable = True

    @torch.no_grad()
    @ignore_numpy_warnings
    def initialize(self):
        super().initialize()

        if self.I0.initialized:
            return

        target_area = self.target[self.window]
        dat = backend.to_numpy(target_area._data).copy()
        mask = backend.to_numpy(target_area._mask)
        dat[mask] = np.median(dat[~mask])

        self.I0.value = np.median(dat) / self.target.pixel_area.item()

    @forward
    def brightness(self, x: ArrayLike, y: ArrayLike, I0: ArrayLike) -> ArrayLike:
        return backend.ones_like(x) * I0



from typing import Union
import torch
import numpy as np

from ..utils.decorators import ignore_numpy_warnings, combine_docstrings
from ..utils.interpolate import interp2d
from .. import config
from .model_object import ComponentModel
from ..backend_obj import backend, ArrayLike
from ..errors import SpecificationConflict
from ..param import forward
from . import func
from ..utils.initialize import polar_decomposition

__all__ = ["BasisModel"]


@combine_docstrings
class BasisModel(ComponentModel):
    """Model described by a set of basis images.

    This model is composed of a set of basis images (think eigen decomposition
    or zernike polynomials) that are linearly combined with some weights to form
    the model image. The basis images are defined on a grid of coordinates,
    and the brightness at any point is determined by bilinear interpolation of
    the basis images. This is a very flexible model that can represent a wide
    range of sources, but depending on the number of basis elements it can become
    computationally expensive to optimize.

    **Parameters:**
        - `weights`: The weights of the basis set of images in units of flux.
        - `PA`: the position angle of the model, in radians.
        - `scale`: the scale of the model, in arcsec per grid unit.
    """

    _model_type = "basis"
    _parameter_specs = {
        "weights": {"units": "unitless", "shape": (None,), "dynamic": True},
        "PA": {"units": "radians", "shape": (), "dynamic": False},
        "scale": {"units": "arcsec/grid-unit", "shape": (), "dynamic": False},
    }
    usable = True

    def __init__(self, *args, basis: Union[str, ArrayLike] = "zernike:3", **kwargs):
        """Initialize the BasisModel with a basis set of images."""
        super().__init__(*args, **kwargs)
        self.basis = basis

    @property
    def basis(self):
        """The basis set of images used to form the model."""
        return self._basis

    @basis.setter
    def basis(self, value: Union[str, ArrayLike]):
        """Set the basis set of images."""
        if value is None:
            raise SpecificationConflict("BasisModel requires a basis set of images to be provided.")
        elif isinstance(value, str) and value.startswith("zernike:"):
            self._basis = value
        else:
            # Transpose since pytorch uses (j, i) indexing when (i, j) is more natural for coordinates
            self._basis = backend.transpose(
                backend.as_array(value, dtype=config.DTYPE, device=config.DEVICE), 2, 1
            )

    @torch.no_grad()
    @ignore_numpy_warnings
    def initialize(self):
        super().initialize()
        if isinstance(self.basis, str) and self.basis.startswith("zernike:"):
            order = int(self.basis.split(":")[1])
            N = int(max(self.window.shape))
            N = N + 1 - N % 2
            self.basis = func.zernike_basis(order, N) / self.target.pixel_area

        if not self.weights.initialized:
            w = np.zeros(self.basis.shape[0])
            w[0] = 1.0
            self.weights.value = w

        if not self.PA.initialized:
            R, _ = polar_decomposition(self.target.CD.npvalue)
            self.PA.value = np.arccos(np.abs(R[0, 0]))

        if not self.scale.initialized:
            self.scale = self.target.pixelscale.item()

    @forward
    def transform_coordinates(
        self, x: ArrayLike, y: ArrayLike, PA: ArrayLike, scale: ArrayLike
    ) -> tuple[ArrayLike, ArrayLike]:
        x, y = super().transform_coordinates(x, y)
        x, y = func.rotate(-PA + np.pi / 2, x, y)
        return x / scale, y / scale

    @forward
    def brightness(self, x: ArrayLike, y: ArrayLike, weights: ArrayLike) -> ArrayLike:
        x, y = self.transform_coordinates(x, y)
        wB = backend.sum(weights[:, None, None] * self.basis, dim=0)
        return interp2d(wB, x + wB.shape[0] // 2, y + wB.shape[1] // 2)



import astrophot as ap
import torch
import numpy as np

import pytest

######################################################################
# Image Objects
######################################################################


@pytest.fixture()
def cmos_target():
    arr = ap.backend.zeros((10, 15))
    return ap.CMOSTargetImage(
        data=arr,
        pixelscale=0.7,
        zeropoint=1.0,
        variance=ap.backend.ones_like(arr),
        mask=ap.backend.zeros_like(arr),
        subpixel_loc=(-0.25, -0.25),
        subpixel_scale=0.5,
    )


def test_cmos_image_creation(cmos_target):
    cmos_copy = cmos_target.copy()
    assert cmos_copy.pixelscale == 0.7, "image should track pixelscale"
    assert cmos_copy.zeropoint == 1.0, "image should track zeropoint"
    assert cmos_copy.crpix[0] == 0, "image should track crpix"
    assert cmos_copy.crpix[1] == 0, "image should track crpix"
    assert cmos_copy.subpixel_loc == (-0.25, -0.25), "image should track subpixel location"
    assert cmos_copy.subpixel_scale == 0.5, "image should track subpixel scale"

    print(cmos_target.data.shape)
    i, j = cmos_target.pixel_center_meshgrid()
    assert i.shape == (15, 10), "meshgrid should have correct shape"
    assert j.shape == (15, 10), "meshgrid should have correct shape"

    x, y = cmos_target.coordinate_center_meshgrid()
    assert x.shape == (15, 10), "coordinate meshgrid should have correct shape"
    assert y.shape == (15, 10), "coordinate meshgrid should have correct shape"


def test_cmos_model_sample(cmos_target):
    model = ap.Model(
        name="test_cmos",
        model_type="sersic galaxy model",
        target=cmos_target,
        center=(3, 5),
        q=0.7,
        PA=np.pi / 3,
        n=2.5,
        Re=4,
        Ie=1.0,
        sampling_mode="midpoint",
        integrate_mode="bright",
    )
    model.initialize()
    img = model()

    assert isinstance(img, ap.CMOSModelImage), "sampled image should be a CMOSModelImage"
    assert img.pixelscale == cmos_target.pixelscale, "sampled image should have the same pixelscale"
    assert img.zeropoint == cmos_target.zeropoint, "sampled image should have the same zeropoint"
    assert (
        img.subpixel_loc == cmos_target.subpixel_loc
    ), "sampled image should have the same subpixel location"


def test_cmos_image_save_load(cmos_target):
    # Save the image
    cmos_target.save("cmos_image.fits")

    # Load the image
    loaded_image = ap.CMOSTargetImage(filename="cmos_image.fits")

    # Check if the loaded image matches the original
    assert ap.backend.allclose(
        cmos_target.data, loaded_image.data
    ), "Loaded image data should match original"
    assert ap.backend.allclose(
        cmos_target.pixelscale, loaded_image.pixelscale
    ), "Loaded image pixelscale should match original"
    assert ap.backend.allclose(
        cmos_target.zeropoint, loaded_image.zeropoint
    ), "Loaded image zeropoint should match original"
    assert np.allclose(
        cmos_target.subpixel_loc, loaded_image.subpixel_loc
    ), "Loaded image subpixel location should match original"
    assert np.allclose(
        cmos_target.subpixel_scale, loaded_image.subpixel_scale
    ), "Loaded image subpixel scale should match original"



import astrophot as ap
import numpy as np


def test_batch_model(sersic):

    M = ap.Model(model_type="batch model", model=sersic)
    assert (
        M.target is sersic.target
    ), "BatchModel should share the same target as its component model"
    assert (
        M.window.extent == sersic.window.extent
    ), "BatchModel should share the same window as its component model"
    assert M.mask is sersic.mask, "BatchModel should share the same mask as its component model"

    sersic.center = [[5, 5], [30, 10], [20, 35]]
    sersic.q = [0.7, 0.4, 0.3]

    gll0 = M.gaussian_log_likelihood()
    pll0 = M.poisson_log_likelihood()
    assert ap.backend.isfinite(gll0), "Gaussian log likelihood should be finite"
    assert ap.backend.isfinite(pll0), "Poisson log likelihood should be finite"
    grad = M.gradient()
    assert ap.backend.all(ap.backend.isfinite(grad)), "Gradient should be finite"
    jac = M.jacobian()
    assert ap.backend.all(ap.backend.isfinite(jac.data)), "Jacobian should be finite"

    res = ap.fit.LM(M, max_iter=5).fit()

    assert len(res.loss_history) >= 2, "Optimizer must be able to find steps to improve the model"
    gll1 = M.gaussian_log_likelihood()
    pll1 = M.poisson_log_likelihood()
    assert ap.backend.isfinite(gll1), "Gaussian log likelihood should be finite"
    assert ap.backend.isfinite(pll1), "Poisson log likelihood should be finite"
    assert gll1 > gll0 and pll1 > pll0, "Model should improve the likelihood after fitting"
    assert np.all(
        np.abs(sersic.q.npvalue - np.array([0.7, 0.4, 0.3])) > 0.1
    ), "Model parameters should change after fitting"
